# Copyright (c) 2014-2016 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Handle binary stream returns in NCStream format."""

from collections import OrderedDict
import itertools
import logging
import zlib

import numpy as np

from . import cdmrfeature_pb2 as cdmrf, ncStream_pb2 as stream

MAGIC_HEADER = b'\xad\xec\xce\xda'
MAGIC_DATA = b'\xab\xec\xce\xba'
MAGIC_DATA2 = b'\xab\xeb\xbe\xba'
MAGIC_VDATA = b'\xab\xef\xfe\xba'
MAGIC_VEND = b'\xed\xef\xfe\xda'
MAGIC_ERR = b'\xab\xad\xba\xda'

MAGIC_HEADERCOV = b'\xad\xed\xde\xda'
MAGIC_DATACOV = b'\xab\xed\xde\xba'

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)


#
# NCStream handling
#
def read_ncstream_data(fobj):
    """Handle reading an NcStream v1 data block from a file-like object."""
    data = read_proto_object(fobj, stream.Data)
    if data.dataType in (stream.STRING, stream.OPAQUE) or data.vdata:
        log.debug('Reading string/opaque/vlen')
        num_obj = read_var_int(fobj)
        log.debug('Num objects: %d', num_obj)
        blocks = [read_block(fobj) for _ in range(num_obj)]
        if data.dataType == stream.STRING:
            blocks = [b.decode('utf-8', errors='ignore') for b in blocks]

        # Again endian isn't coded properly
        dt = data_type_to_numpy(data.dataType).newbyteorder('>')
        if data.vdata:
            return np.array([np.frombuffer(b, dtype=dt) for b in blocks], dtype=object)
        else:
            return np.array(blocks, dtype=dt)
    elif data.dataType in _dtype_lookup:
        log.debug('Reading array data')
        bin_data = read_block(fobj)
        log.debug('Binary data: %s', bin_data)

        # Hard code to big endian for now since it's not encoded correctly
        dt = data_type_to_numpy(data.dataType).newbyteorder('>')

        # Handle decompressing the bytes
        if data.compress == stream.DEFLATE:
            bin_data = zlib.decompress(bin_data)
            if len(bin_data) != data.uncompressedSize:
                log.error('Uncompressed size mismatch %d vs. %d', len(bin_data),
                          data.uncompressedSize)
        elif data.compress != stream.NONE:
            raise NotImplementedError(f'Compression type {data.compress} not implemented!')

        # Turn bytes into an array
        return reshape_array(data, np.frombuffer(bin_data, dtype=dt))
    elif data.dataType == stream.STRUCTURE:
        sd = read_proto_object(fobj, stream.StructureData)

        # Make a datatype appropriate to the rows of struct
        endian = '>' if data.bigend else '<'
        dt = np.dtype([(endian, np.void, sd.rowLength)])

        # Turn bytes into an array
        return reshape_array(data, np.frombuffer(sd.data, dtype=dt))
    elif data.dataType == stream.SEQUENCE:
        log.debug('Reading sequence')
        blocks = []
        magic = read_magic(fobj)
        while magic != MAGIC_VEND:
            if magic == MAGIC_VDATA:
                log.error('Bad magic for struct/seq data!')
            blocks.append(read_proto_object(fobj, stream.StructureData))
            magic = read_magic(fobj)
        return data, blocks
    else:
        raise NotImplementedError(f"Don't know how to handle data type: {data.dataType}")


def read_ncstream_data2(fobj):
    """Handle reading an NcStream v2 data block from a file-like object."""
    data = read_proto_object(fobj, stream.DataCol)
    return datacol_to_array(data)


def read_ncstream_err(fobj):
    """Handle reading an NcStream error from a file-like object and raise as error."""
    err = read_proto_object(fobj, stream.Error)
    raise RuntimeError(err.message)


ncstream_table = {MAGIC_HEADER: lambda f: read_proto_object(f, stream.Header),
                  MAGIC_DATA: read_ncstream_data,
                  MAGIC_DATA2: read_ncstream_data2,
                  MAGIC_ERR: read_ncstream_err}


def read_ncstream_messages(fobj):
    """Read a collection of NcStream messages from a file-like object."""
    return read_messages(fobj, ncstream_table)


#
# CDMRemoteFeature handling
#
cdmrf_table = {MAGIC_HEADERCOV: lambda f: read_proto_object(f, cdmrf.CoverageDataset),
               MAGIC_DATACOV: lambda f: read_proto_object(f, cdmrf.CoverageDataResponse),
               MAGIC_DATA2: read_ncstream_data2,  # For coordinates
               MAGIC_ERR: read_ncstream_err}


def read_cdmrf_messages(fobj):
    """Read a collection of CDMRemoteFeature messages from a file-like object."""
    return read_messages(fobj, cdmrf_table)


#
# General Utilities
#
def read_messages(fobj, magic_table):
    """Read messages from a file-like object until stream is exhausted."""
    messages = []

    while True:
        magic = read_magic(fobj)
        if not magic:
            break

        func = magic_table.get(magic)
        if func is not None:
            messages.append(func(fobj))
        else:
            log.error('Unknown magic: %s', ' '.join(f'{b: 02x}' for b in bytearray(magic)))

    return messages


def read_proto_object(fobj, klass):
    """Read a block of data and parse using the given protobuf object."""
    log.debug('%s chunk', klass.__name__)
    obj = klass()
    obj.ParseFromString(read_block(fobj))
    log.debug('Header: %s', str(obj))
    return obj


def read_magic(fobj):
    """Read magic bytes.

    Parameters
    ----------
    fobj : file-like object
        The file to read from.

    Returns
    -------
    bytes
        magic byte sequence read

    """
    return fobj.read(4)


def read_block(fobj):
    """Read a block.

    Reads a block from a file object by first reading the number of bytes to read, which must
    be encoded as a variable-byte length integer.

    Parameters
    ----------
    fobj : file-like object
        The file to read from.

    Returns
    -------
    bytes
        block of bytes read

    """
    num = read_var_int(fobj)
    log.debug('Next block: %d bytes', num)
    return fobj.read(num)


def process_vlen(data_header, array):
    """Process vlen coming back from NCStream v2.

    This takes the array of values and slices into an object array, with entries containing
    the appropriate pieces of the original array. Sizes are controlled by the passed in
    `data_header`.

    Parameters
    ----------
    data_header : Header
    array : :class:`numpy.ndarray`

    Returns
    -------
    ndarray
        object array containing sub-sequences from the original primitive array

    """
    source = iter(array)
    return np.array([np.fromiter(itertools.islice(source, size), dtype=array.dtype)
                     for size in data_header.vlens], dtype=object)


def datacol_to_array(datacol):
    """Convert DataCol from NCStream v2 into an array with appropriate type.

    Depending on the data type specified, this extracts data from the appropriate members
    and packs into a :class:`numpy.ndarray`, recursing as necessary for compound data types.

    Parameters
    ----------
    datacol : DataCol

    Returns
    -------
    ndarray
        array containing extracted data

    """
    if datacol.dataType == stream.STRING:
        arr = np.array(datacol.stringdata, dtype=object)
    elif datacol.dataType == stream.OPAQUE:
        arr = np.array(datacol.opaquedata, dtype=object)
    elif datacol.dataType == stream.STRUCTURE:
        members = OrderedDict((mem.name, datacol_to_array(mem))
                              for mem in datacol.structdata.memberData)
        log.debug('Struct members:\n%s', str(members))

        # str() around name necessary because protobuf gives unicode names, but dtype doesn't
        # support them on Python 2
        dt = np.dtype([(str(name), arr.dtype) for name, arr in members.items()])
        log.debug('Struct dtype: %s', str(dt))

        arr = np.empty((datacol.nelems,), dtype=dt)
        for name, arr_data in members.items():
            arr[name] = arr_data
    else:
        # Make an appropriate datatype
        endian = '>' if datacol.bigend else '<'
        dt = data_type_to_numpy(datacol.dataType).newbyteorder(endian)

        # Turn bytes into an array
        arr = np.frombuffer(datacol.primdata, dtype=dt)
        if arr.size != datacol.nelems:
            log.warning('Array size %d does not agree with nelems %d',
                        arr.size, datacol.nelems)
        if datacol.isVlen:
            arr = process_vlen(datacol, arr)
            try:
                arr = reshape_array(datacol, arr)
            except ValueError:
                # In this case, the array collapsed, need different resize that
                # correctly sizes from elements
                shape = tuple(r.size for r in datacol.section.range) + (datacol.vlens[0],)
                arr = arr.reshape(*shape)
        else:
            arr = reshape_array(datacol, arr)
    return arr


def reshape_array(data_header, array):
    """Extract the appropriate array shape from the header.

    Can handle taking a data header and either bytes containing data or a StructureData
    instance, which will have binary data as well as some additional information.

    Parameters
    ----------
    array : :class:`numpy.ndarray`
    data_header : Data

    """
    shape = tuple(r.size for r in data_header.section.range)
    if shape:
        return array.reshape(*shape)
    else:
        return array


# STRUCTURE = 8;
# SEQUENCE = 9;
_dtype_lookup = {stream.CHAR: 'S1', stream.BYTE: 'b', stream.SHORT: 'i2',
                 stream.INT: 'i4', stream.LONG: 'i8', stream.FLOAT: 'f4',
                 stream.DOUBLE: 'f8', stream.STRING: 'O',
                 stream.ENUM1: 'B', stream.ENUM2: 'u2', stream.ENUM4: 'u4',
                 stream.OPAQUE: 'O', stream.UBYTE: 'B', stream.USHORT: 'u2',
                 stream.UINT: 'u4', stream.ULONG: 'u8'}


def data_type_to_numpy(datatype, unsigned=False):
    """Convert an ncstream datatype to a numpy one."""
    basic_type = _dtype_lookup[datatype]

    if datatype in (stream.STRING, stream.OPAQUE):
        return np.dtype(basic_type)

    if unsigned:
        basic_type = basic_type.replace('i', 'u')
    return np.dtype('=' + basic_type)


def struct_to_dtype(struct):
    """Convert a Structure specification to a numpy structured dtype."""
    # str() around name necessary because protobuf gives unicode names, but dtype doesn't
    # support them on Python 2
    fields = [(str(var.name), data_type_to_numpy(var.dataType, var.unsigned))
              for var in struct.vars]
    for s in struct.structs:
        fields.append((str(s.name), struct_to_dtype(s)))

    log.debug('Structure fields: %s', fields)
    dt = np.dtype(fields)
    return dt


def unpack_variable(var):
    """Unpack an NCStream Variable into information we can use."""
    # If we actually get a structure instance, handle turning that into a variable
    if var.dataType == stream.STRUCTURE:
        return None, struct_to_dtype(var), 'Structure'
    elif var.dataType == stream.SEQUENCE:
        log.warning('Sequence support not implemented!')

    dt = data_type_to_numpy(var.dataType, var.unsigned)
    if var.dataType == stream.OPAQUE:
        type_name = 'opaque'
    elif var.dataType == stream.STRING:
        type_name = 'string'
    else:
        type_name = dt.name

    if var.data:
        log.debug('Storing variable data: %s %s', dt, var.data)
        if var.dataType == stream.STRING:
            data = var.data
        else:
            # Always sent big endian
            data = np.frombuffer(var.data, dtype=dt.newbyteorder('>'))
    else:
        data = None

    return data, dt, type_name


_attr_converters = {stream.Attribute.BYTE: np.dtype('>b'),
                    stream.Attribute.SHORT: np.dtype('>i2'),
                    stream.Attribute.INT: np.dtype('>i4'),
                    stream.Attribute.LONG: np.dtype('>i8'),
                    stream.Attribute.FLOAT: np.dtype('>f4'),
                    stream.Attribute.DOUBLE: np.dtype('>f8')}


def unpack_attribute(att):
    """Unpack an embedded attribute into a python or numpy object."""
    if att.unsigned:
        log.warning('Unsupported unsigned attribute!')

    # TDS 5.0 now has a dataType attribute that takes precedence
    if att.len == 0:  # Empty
        val = None
    elif att.dataType == stream.STRING:  # Then look for new datatype string
        val = att.sdata
    elif att.dataType:  # Then a non-zero new data type
        val = np.frombuffer(att.data,
                            dtype='>' + _dtype_lookup[att.dataType], count=att.len)
    elif att.type:  # Then non-zero old-data type0
        val = np.frombuffer(att.data,
                            dtype=_attr_converters[att.type], count=att.len)
    elif att.sdata:  # This leaves both 0, try old string
        val = att.sdata
    else:  # Assume new datatype is Char (0)
        val = np.array(att.data, dtype=_dtype_lookup[att.dataType])

    if att.len == 1:
        val = val[0]

    return att.name, val


def read_var_int(file_obj):
    """Read a variable-length integer.

    Parameters
    ----------
    file_obj : file-like object
        The file to read from.

    Returns
    -------
    int
        the variable-length value read

    """
    # Read all bytes from here, stopping with the first one that does not have
    # the MSB set. Save the lower 7 bits, and keep stacking to the *left*.
    val = 0
    shift = 0
    while True:
        # Read next byte
        next_val = ord(file_obj.read(1))
        val |= ((next_val & 0x7F) << shift)
        shift += 7
        if not next_val & 0x80:
            break

    return val



# Copyright (c) 2021 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Tools for versioning."""


def get_version():
    """Get Siphon's version.

    Either get it from package metadata, or get it using version control information if
    a development install.
    """
    try:
        from setuptools_scm import get_version as _get_version
        return _get_version(root='../..', relative_to=__file__,
                            version_scheme='post-release')
    except (ImportError, LookupError):
        from importlib.metadata import PackageNotFoundError, version

        try:
            return version(__package__)
        except PackageNotFoundError:
            return 'Unknown'



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Contains functionality for retrieving simple web service data."""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ncStream.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0encStream.proto\"\xe4\x01\n\tAttribute\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1d\n\x04type\x18\x02 \x01(\x0e\x32\x0f.Attribute.Type\x12\x0b\n\x03len\x18\x03 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\r\n\x05sdata\x18\x05 \x03(\t\x12\x10\n\x08unsigned\x18\x06 \x01(\x08\x12\x1b\n\x08\x64\x61taType\x18\x07 \x01(\x0e\x32\t.DataType\"Q\n\x04Type\x12\n\n\x06STRING\x10\x00\x12\x08\n\x04\x42YTE\x10\x01\x12\t\n\x05SHORT\x10\x02\x12\x07\n\x03INT\x10\x03\x12\x08\n\x04LONG\x10\x04\x12\t\n\x05\x46LOAT\x10\x05\x12\n\n\x06\x44OUBLE\x10\x06\"a\n\tDimension\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06length\x18\x02 \x01(\x04\x12\x13\n\x0bisUnlimited\x18\x03 \x01(\x08\x12\x0e\n\x06isVlen\x18\x04 \x01(\x08\x12\x11\n\tisPrivate\x18\x05 \x01(\x08\"\x9c\x01\n\x08Variable\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\x19\n\x05shape\x18\x03 \x03(\x0b\x32\n.Dimension\x12\x18\n\x04\x61tts\x18\x04 \x03(\x0b\x32\n.Attribute\x12\x10\n\x08unsigned\x18\x05 \x01(\x08\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\x12\x10\n\x08\x65numType\x18\x07 \x01(\t\"\xa1\x01\n\tStructure\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\x19\n\x05shape\x18\x03 \x03(\x0b\x32\n.Dimension\x12\x18\n\x04\x61tts\x18\x04 \x03(\x0b\x32\n.Attribute\x12\x17\n\x04vars\x18\x05 \x03(\x0b\x32\t.Variable\x12\x1b\n\x07structs\x18\x06 \x03(\x0b\x32\n.Structure\"h\n\x0b\x45numTypedef\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\"\n\x03map\x18\x02 \x03(\x0b\x32\x15.EnumTypedef.EnumType\x1a\'\n\x08\x45numType\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\t\"\xb8\x01\n\x05Group\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x04\x64ims\x18\x02 \x03(\x0b\x32\n.Dimension\x12\x17\n\x04vars\x18\x03 \x03(\x0b\x32\t.Variable\x12\x1b\n\x07structs\x18\x04 \x03(\x0b\x32\n.Structure\x12\x18\n\x04\x61tts\x18\x05 \x03(\x0b\x32\n.Attribute\x12\x16\n\x06groups\x18\x06 \x03(\x0b\x32\x06.Group\x12\x1f\n\tenumTypes\x18\x07 \x03(\x0b\x32\x0c.EnumTypedef\"\\\n\x06Header\x12\x10\n\x08location\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12\x14\n\x04root\x18\x04 \x01(\x0b\x32\x06.Group\x12\x0f\n\x07version\x18\x05 \x01(\r\"&\n\x05\x45rror\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\r\"4\n\x05Range\x12\r\n\x05start\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\x04\x12\x0e\n\x06stride\x18\x03 \x01(\x04\" \n\x07Section\x12\x15\n\x05range\x18\x01 \x03(\x0b\x32\x06.Range\"\xca\x01\n\x04\x44\x61ta\x12\x0f\n\x07varName\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\x19\n\x07section\x18\x03 \x01(\x0b\x32\x08.Section\x12\x10\n\x06\x62igend\x18\x04 \x01(\x08H\x00\x12\x0f\n\x07version\x18\x05 \x01(\r\x12\x1b\n\x08\x63ompress\x18\x06 \x01(\x0e\x32\t.Compress\x12\r\n\x05vdata\x18\x07 \x01(\x08\x12\x18\n\x10uncompressedSize\x18\x08 \x01(\rB\x10\n\x0e\x62igend_present\"q\n\rStructureData\x12\x0e\n\x06member\x18\x01 \x03(\r\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x11\n\theapCount\x18\x03 \x03(\r\x12\r\n\x05sdata\x18\x04 \x03(\t\x12\r\n\x05nrows\x18\x05 \x01(\x04\x12\x11\n\trowLength\x18\x06 \x01(\r\"\x81\x02\n\x07\x44\x61taCol\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\x19\n\x07section\x18\x03 \x01(\x0b\x32\x08.Section\x12\x0e\n\x06\x62igend\x18\x04 \x01(\x08\x12\x0f\n\x07version\x18\x05 \x01(\r\x12\x0e\n\x06isVlen\x18\x07 \x01(\x08\x12\x0e\n\x06nelems\x18\t \x01(\r\x12\x10\n\x08primdata\x18\n \x01(\x0c\x12\x12\n\nstringdata\x18\x0b \x03(\t\x12\r\n\x05vlens\x18\x0c \x03(\r\x12\x12\n\nopaquedata\x18\r \x03(\x0c\x12&\n\nstructdata\x18\x0e \x01(\x0b\x32\x12.ArrayStructureCol\"1\n\x11\x41rrayStructureCol\x12\x1c\n\nmemberData\x18\x01 \x03(\x0b\x32\x08.DataCol\"\x85\x02\n\x07\x44\x61taRow\x12\x10\n\x08\x66ullName\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\x19\n\x07section\x18\x03 \x01(\x0b\x32\x08.Section\x12\x0e\n\x06\x62igend\x18\x04 \x01(\x08\x12\x0f\n\x07version\x18\x05 \x01(\r\x12\x0e\n\x06isVlen\x18\x07 \x01(\x08\x12\x0e\n\x06nelems\x18\t \x01(\r\x12\x10\n\x08primdata\x18\n \x01(\x0c\x12\x12\n\nstringdata\x18\x0b \x03(\t\x12\r\n\x05vlens\x18\x0c \x03(\r\x12\x12\n\nopaquedata\x18\r \x03(\x0c\x12&\n\nstructdata\x18\x0e \x01(\x0b\x32\x12.ArrayStructureRow\"W\n\x06Member\x12\x11\n\tshortName\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\r\n\x05shape\x18\x03 \x03(\r\x12\x0e\n\x06isVlen\x18\x04 \x01(\x08\"\xae\x01\n\x11\x41rrayStructureRow\x12\x18\n\x07members\x18\x01 \x03(\x0b\x32\x07.Member\x12\r\n\x05nrows\x18\x05 \x01(\x04\x12\x11\n\trowLength\x18\x06 \x01(\r\x12\x0f\n\x07\x66ixdata\x18\n \x01(\x0c\x12\x12\n\nstringdata\x18\x0b \x03(\t\x12\x10\n\x08\x62ytedata\x18\r \x03(\x0c\x12&\n\nstructdata\x18\x0e \x03(\x0b\x32\x12.ArrayStructureRow*\xd5\x01\n\x08\x44\x61taType\x12\x08\n\x04\x43HAR\x10\x00\x12\x08\n\x04\x42YTE\x10\x01\x12\t\n\x05SHORT\x10\x02\x12\x07\n\x03INT\x10\x03\x12\x08\n\x04LONG\x10\x04\x12\t\n\x05\x46LOAT\x10\x05\x12\n\n\x06\x44OUBLE\x10\x06\x12\n\n\x06STRING\x10\x07\x12\r\n\tSTRUCTURE\x10\x08\x12\x0c\n\x08SEQUENCE\x10\t\x12\t\n\x05\x45NUM1\x10\n\x12\t\n\x05\x45NUM2\x10\x0b\x12\t\n\x05\x45NUM4\x10\x0c\x12\n\n\x06OPAQUE\x10\r\x12\t\n\x05UBYTE\x10\x0e\x12\n\n\x06USHORT\x10\x0f\x12\x08\n\x04UINT\x10\x10\x12\t\n\x05ULONG\x10\x11*!\n\x08\x43ompress\x12\x08\n\x04NONE\x10\x00\x12\x0b\n\x07\x44\x45\x46LATE\x10\x01\x42 \n\x0fucar.nc2.streamB\rNcStreamProtob\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ncStream_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017ucar.nc2.streamB\rNcStreamProto'
  _DATATYPE._serialized_start=2348
  _DATATYPE._serialized_end=2561
  _COMPRESS._serialized_start=2563
  _COMPRESS._serialized_end=2596
  _ATTRIBUTE._serialized_start=19
  _ATTRIBUTE._serialized_end=247
  _ATTRIBUTE_TYPE._serialized_start=166
  _ATTRIBUTE_TYPE._serialized_end=247
  _DIMENSION._serialized_start=249
  _DIMENSION._serialized_end=346
  _VARIABLE._serialized_start=349
  _VARIABLE._serialized_end=505
  _STRUCTURE._serialized_start=508
  _STRUCTURE._serialized_end=669
  _ENUMTYPEDEF._serialized_start=671
  _ENUMTYPEDEF._serialized_end=775
  _ENUMTYPEDEF_ENUMTYPE._serialized_start=736
  _ENUMTYPEDEF_ENUMTYPE._serialized_end=775
  _GROUP._serialized_start=778
  _GROUP._serialized_end=962
  _HEADER._serialized_start=964
  _HEADER._serialized_end=1056
  _ERROR._serialized_start=1058
  _ERROR._serialized_end=1096
  _RANGE._serialized_start=1098
  _RANGE._serialized_end=1150
  _SECTION._serialized_start=1152
  _SECTION._serialized_end=1184
  _DATA._serialized_start=1187
  _DATA._serialized_end=1389
  _STRUCTUREDATA._serialized_start=1391
  _STRUCTUREDATA._serialized_end=1504
  _DATACOL._serialized_start=1507
  _DATACOL._serialized_end=1764
  _ARRAYSTRUCTURECOL._serialized_start=1766
  _ARRAYSTRUCTURECOL._serialized_end=1815
  _DATAROW._serialized_start=1818
  _DATAROW._serialized_end=2079
  _MEMBER._serialized_start=2081
  _MEMBER._serialized_end=2168
  _ARRAYSTRUCTUREROW._serialized_start=2171
  _ARRAYSTRUCTUREROW._serialized_end=2345
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Read upper air data from the Wyoming archives."""

from datetime import datetime
from io import StringIO
import warnings

from bs4 import BeautifulSoup
import numpy as np
import pandas as pd

from .._tools import get_wind_components
from ..http_util import HTTPEndPoint


def _safe_float(s):
    """Convert to float, handling ****** as a string for missing."""
    return pd.NA if all(c == '*' for c in s) or s == '-9999.0' else float(s)


class WyomingUpperAir(HTTPEndPoint):
    """Download and parse data from the University of Wyoming's upper air archive."""

    def __init__(self):
        """Set up endpoint."""
        super().__init__('http://weather.uwyo.edu/cgi-bin/sounding')

    @classmethod
    def request_data(cls, time, site_id, recalc=False, **kwargs):
        r"""Retrieve upper air observations from the Wyoming archive.

        Parameters
        ----------
        time : datetime.datetime
            The date and time of the desired observation.

        site_id : str
            The three letter ICAO identifier of the station for which data should be
            downloaded.

        recalc : bool
            Whether to request the server recalculate the data (i.e. ignore its cache) before
            returning it. Defaults to False. NOTE: This should be used sparingly because it
            will increase the load on the service.

        kwargs
            Arbitrary keyword arguments to use to initialize source

        Returns
        -------
        `pandas.DataFrame`
            Parsed data

        """
        endpoint = cls()
        df = endpoint._get_data(time, site_id, recalc=recalc)
        return df

    def _get_data(self, time, site_id, recalc=False):
        r"""Download and parse upper air observations from an online archive.

        Parameters
        ----------
        time : datetime.datetime
            The date and time of the desired observation.

        site_id : str
            The three letter ICAO identifier of the station for which data should be
            downloaded.

        recalc : bool
            Returns recalculated data if True. Defaults to False.

        Returns
        -------
        `pandas.DataFrame`

        """
        raw_data = self._get_data_raw(time, site_id, recalc=recalc)
        soup = BeautifulSoup(raw_data, 'html.parser')
        tabular_data = StringIO(soup.find_all('pre')[0].contents[0])
        col_names = ['pressure', 'height', 'temperature', 'dewpoint', 'direction', 'speed']
        df = pd.read_fwf(tabular_data, widths=[7] * 8, skiprows=5,
                         usecols=[0, 1, 2, 3, 6, 7], names=col_names)

        df['u_wind'], df['v_wind'] = get_wind_components(df['speed'],
                                                         np.deg2rad(df['direction']))

        # Drop any rows with all NaN values for T, Td, winds
        df = df.dropna(subset=('temperature', 'dewpoint', 'direction', 'speed',
                               'u_wind', 'v_wind'), how='all').reset_index(drop=True)

        # Parse metadata
        meta_data = soup.find_all('pre')[1].contents[0]
        lines = meta_data.splitlines()

        # Convert values after table into key, value pairs using the name to the left of the :
        post_values = dict(tuple(map(str.strip, line.split(': '))) for line in lines[1:])

        station = post_values.get('Station identifier', '')
        station_number = int(post_values['Station number'])
        sounding_time = datetime.strptime(post_values['Observation time'], '%y%m%d/%H%M')
        latitude = _safe_float(post_values.get('Station latitude', '******'))
        longitude = _safe_float(post_values.get('Station longitude', '******'))
        elevation = _safe_float(post_values.get('Station elevation', '******'))
        pw = float(post_values['Precipitable water [mm] for entire sounding'])

        df['station'] = station
        df['station_number'] = station_number
        df['time'] = sounding_time
        df['latitude'] = latitude
        df['longitude'] = longitude
        df['elevation'] = elevation
        df['pw'] = pw

        # Add unit dictionary
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = {'pressure': 'hPa',
                        'height': 'meter',
                        'temperature': 'degC',
                        'dewpoint': 'degC',
                        'direction': 'degrees',
                        'speed': 'knot',
                        'u_wind': 'knot',
                        'v_wind': 'knot',
                        'station': None,
                        'station_number': None,
                        'time': None,
                        'latitude': 'degrees',
                        'longitude': 'degrees',
                        'elevation': 'meter',
                        'pw': 'millimeter'}

        return df

    def _get_data_raw(self, time, site_id, recalc=False):
        """Download data from the University of Wyoming's upper air archive.

        Parameters
        ----------
        time : datetime.datetime
            Date and time for which data should be downloaded
        site_id : str
            Site id for which data should be downloaded
        recalc : bool
            Returns recalculated data if True. Defaults to False.

        Returns
        -------
        str
            text of the server response

        """
        path = ('?region=naconf&TYPE=TEXT%3ALIST'
                f'&YEAR={time:%Y}&MONTH={time:%m}&FROM={time:%d%H}&TO={time:%d%H}'
                f'&STNM={site_id}')
        if recalc:
            path += '&REPLOT=1'

        resp = self.get_path(path)
        # See if the return is valid, but has no data
        if resp.text.find("Can't") != -1:
            raise ValueError(
                f'No data available for {time:%Y-%m-%d %HZ} '
                f'for station {site_id}.')

        return resp.text



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Tools for data conversion and presentation."""

import numpy as np


def get_wind_components(speed, wdir):
    r"""Calculate the U, V wind vector components from the speed and direction.

    Parameters
    ----------
    speed : array_like
        The wind speed (magnitude)
    wdir : array_like
        The wind direction, specified as the direction from which the wind is
        blowing, with 0 being North.

    Returns
    -------
    u, v : tuple of array_like
        The wind components in the X (East-West) and Y (North-South)
        directions, respectively.

    """
    u = -speed * np.sin(wdir)
    v = -speed * np.cos(wdir)
    return u, v



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Utilities for testing siphon."""

import os.path

import vcr


def get_recorder(test_file_path):
    """Return an appropriate response recorder for the given path."""
    return vcr.VCR(cassette_library_dir=os.path.join(os.path.dirname(test_file_path),
                                                     'fixtures'))



# Copyright (c) 2016 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Provide access to a TDS Coverage Dataset."""

from collections import OrderedDict
import logging
import warnings

from .cdmremotefeature import CDMRemoteFeature
from .dataset import AttributeContainer

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)


def reindent_lines(new_leader, source):
    """Re-indent string lines."""
    return new_leader + ('\n' + new_leader).join(source.split('\n'))


class CoverageDataset(AttributeContainer):
    """Wrap dataset access using CDMRemoteFeature and Coverages.

    This is still experimental.
    """

    def __init__(self, url):
        """Initialize CoverageDataset from a url pointing to CDMRemoteFeature endpoint."""
        super().__init__()
        warnings.warn('CoverageDataset is in early development, unsupported, and API may '
                      'change at any time.', stacklevel=2)
        self.cdmrf = CDMRemoteFeature(url)
        self.name = 'Unnamed'
        self.lon_lat_domain = None
        self.proj_domain = None
        self.date_range = None
        self.type = None
        self.axes = OrderedDict()
        self.coord_systems = OrderedDict()
        self.grids = OrderedDict()
        self.transforms = OrderedDict()
        self._read_header()

    def _read_header(self):
        """Get the needed header information to initialize dataset."""
        self._header = self.cdmrf.fetch_header()
        self.load_from_stream(self._header)

    def load_from_stream(self, header):
        """Populate the CoverageDataset from the protobuf information."""
        self._unpack_attrs(header.atts)
        self.name = header.name
        self.lon_lat_domain = header.latlonRect
        self.proj_domain = header.projRect
        self.date_range = header.dateRange
        self.type = header.coverageType

        for sys in header.coordSys:
            self.coord_systems[sys.name] = sys

        for trans in header.coordTransforms:
            self.transforms[trans.name] = trans

        for ax in header.coordAxes:
            self.axes[ax.name] = ax

        for cov in header.grids:
            self.grids[cov.name] = cov

    def __str__(self):
        """Create a string representation of CoverageDataset."""
        print_groups = []
        if self.name:
            print_groups.append(self.name + ' (' + str(self.type) + ')')

        print_groups.append(f'Lon/Lat Domain: {self.lon_lat_domain}')
        print_groups.append(f'Projected Domain: {self.proj_domain}')
        print_groups.append(f'Date Range: {self.date_range}')

        indent = ' ' * 4
        if self.axes:
            print_groups.append('Axes:')
            for ax in self.axes.values():
                print_groups.append(reindent_lines(indent, str(ax)))

        if self.coord_systems:
            print_groups.append('Coordinate Systems:')
            for sys in self.coord_systems.values():
                print_groups.append(reindent_lines(indent, str(sys)))

        if self.transforms:
            print_groups.append('Coordinate Transforms:')
            for trans in self.transforms.values():
                print_groups.append(reindent_lines(indent, str(trans)))

        if self.grids:
            print_groups.append('Grids:')
            for grid in self.grids.values():
                print_groups.append(reindent_lines(indent, str(grid)))

        if self.ncattrs():
            print_groups.append('Attributes:')
            for att in self.ncattrs():
                print_groups.append(f'{indent}{att}: {getattr(self, att)}')
        return '\n'.join(print_groups)



# Copyright (c) 2013-2021 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Tools for accessing atmospheric and oceanic science data on remote servers."""

# Version import needs to come first so everyone else can pull on import
from ._version import get_version

__version__ = get_version()
del get_version



# Copyright (c) 2016,2019 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Implement an experimental backend for using xarray to talk to TDS over CDMRemote."""

from xarray import Variable
from xarray.backends.common import AbstractDataStore, BackendArray
from xarray.core import indexing
from xarray.core.utils import FrozenDict

from . import Dataset


class CDMArrayWrapper(BackendArray):
    """Wrap a CDMRemote variable for access by xarray."""

    def __init__(self, variable_name, datastore):
        """Initialize the wrapper."""
        self.datastore = datastore
        self.variable_name = variable_name

        array = self.get_array()
        self.shape = array.shape
        self.dtype = array.dtype

    def get_array(self):
        """Get the actual array data from CDM Remote."""
        return self.datastore.ds.variables[self.variable_name]

    def __getitem__(self, item):
        """Wrap getitem around the data."""
        with self.datastore:
            arr = self.get_array()
            return indexing.explicit_indexing_adapter(item, self.shape,
                                                      indexing.IndexingSupport.BASIC,
                                                      arr.__getitem__)


class CDMRemoteStore(AbstractDataStore):
    """Manage a store for accessing CDMRemote datasets with Siphon."""

    def __init__(self, url, deflate=None):
        """Initialize the data store."""
        self.ds = Dataset(url)
        if deflate is not None:
            self.ds.cdmr.deflate = deflate

    def open_store_variable(self, name, var):
        """Turn CDMRemote variable into something like a numpy.ndarray."""
        data = indexing.LazilyOuterIndexedArray(CDMArrayWrapper(name, self))
        return Variable(var.dimensions, data, {a: getattr(var, a) for a in var.ncattrs()})

    def get_variables(self):
        """Get the variables from underlying data set."""
        return FrozenDict((k, self.open_store_variable(k, v))
                          for k, v in self.ds.variables.items())

    def get_attrs(self):
        """Get the global attributes from underlying data set."""
        return FrozenDict((a, getattr(self.ds, a)) for a in self.ds.ncattrs())

    def get_dimensions(self):
        """Get the dimensions from underlying data set."""
        return FrozenDict((k, len(v)) for k, v in self.ds.dimensions.items())



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Support for using CDMRemote on a THREDDS Data Server (TDS)."""

from .dataset import Dataset

__all__ = ['Dataset']



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cdmrfeature.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# Manually added package-relative import
from . import ncStream_pb2 as ncStream__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x63\x64mrfeature.proto\x1a\x0encStream.proto\"K\n\x0e\x43oordTransform\x12\x0f\n\x07isHoriz\x18\x01 \x01(\x08\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1a\n\x06params\x18\x03 \x03(\x0b\x32\n.Attribute\"\xce\x02\n\tCoordAxis\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\x18\n\x04\x61tts\x18\x03 \x03(\x0b\x32\n.Attribute\x12\x1b\n\x08\x61xisType\x18\x04 \x01(\x0e\x32\t.AxisType\x12\r\n\x05units\x18\x05 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x06 \x01(\t\x12\x1f\n\x06\x64\x65pend\x18\x07 \x01(\x0e\x32\x0f.DependenceType\x12\x11\n\tdependsOn\x18\x08 \x03(\t\x12\r\n\x05shape\x18\t \x03(\r\x12\x0f\n\x07nvalues\x18\n \x01(\x03\x12\x1d\n\x07spacing\x18\x0b \x01(\x0e\x32\x0c.AxisSpacing\x12\x12\n\nstartValue\x18\x0c \x01(\x01\x12\x10\n\x08\x65ndValue\x18\r \x01(\x01\x12\x12\n\nresolution\x18\x0e \x01(\x01\x12\x0e\n\x06values\x18\x0f \x01(\x0c\"h\n\x08\x43oordSys\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\taxisNames\x18\x02 \x03(\t\x12\x16\n\x0etransformNames\x18\x03 \x03(\t\x12#\n\x0c\x63overageType\x18\x05 \x01(\x0e\x32\r.CoverageType\"L\n\x11\x43\x61lendarDateRange\x12\r\n\x05start\x18\x01 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x03\x12\x1b\n\x08\x63\x61lendar\x18\x03 \x01(\x0e\x32\t.Calendar\"G\n\tRectangle\x12\x0e\n\x06startx\x18\x01 \x01(\x01\x12\x0e\n\x06starty\x18\x02 \x01(\x01\x12\x0c\n\x04incx\x18\x03 \x01(\x01\x12\x0c\n\x04incy\x18\x04 \x01(\x01\"\x85\x01\n\x08\x43overage\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\x18\n\x04\x61tts\x18\x04 \x03(\x0b\x32\n.Attribute\x12\x10\n\x08\x63oordSys\x18\x05 \x01(\t\x12\r\n\x05units\x18\x06 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x07 \x01(\t\"\xc3\x02\n\x0f\x43overageDataset\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x04\x61tts\x18\x02 \x03(\x0b\x32\n.Attribute\x12\x1e\n\nlatlonRect\x18\x03 \x01(\x0b\x32\n.Rectangle\x12\x1c\n\x08projRect\x18\x04 \x01(\x0b\x32\n.Rectangle\x12%\n\tdateRange\x18\x05 \x01(\x0b\x32\x12.CalendarDateRange\x12\x1b\n\x08\x63oordSys\x18\x06 \x03(\x0b\x32\t.CoordSys\x12(\n\x0f\x63oordTransforms\x18\x07 \x03(\x0b\x32\x0f.CoordTransform\x12\x1d\n\tcoordAxes\x18\x08 \x03(\x0b\x32\n.CoordAxis\x12\x18\n\x05grids\x18\t \x03(\x0b\x32\t.Coverage\x12#\n\x0c\x63overageType\x18\n \x01(\x0e\x32\r.CoverageType\"\xe8\x01\n\x12GeoReferencedArray\x12\x14\n\x0c\x63overageName\x18\x01 \x01(\t\x12\x1b\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\t.DataType\x12\x0e\n\x06\x62igend\x18\x03 \x01(\x08\x12\x0f\n\x07version\x18\x04 \x01(\r\x12\x1b\n\x08\x63ompress\x18\x05 \x01(\x0e\x32\t.Compress\x12\x18\n\x10uncompressedSize\x18\x06 \x01(\x04\x12\r\n\x05shape\x18\x07 \x03(\r\x12\x10\n\x08\x61xisName\x18\x08 \x03(\t\x12\x14\n\x0c\x63oordSysName\x18\t \x01(\t\x12\x10\n\x08primdata\x18\n \x01(\x0c\"\xa3\x01\n\x14\x43overageDataResponse\x12\x1d\n\tcoordAxes\x18\x01 \x03(\x0b\x32\n.CoordAxis\x12\x1b\n\x08\x63oordSys\x18\x02 \x03(\x0b\x32\t.CoordSys\x12(\n\x0f\x63oordTransforms\x18\x03 \x03(\x0b\x32\x0f.CoordTransform\x12%\n\x08geoArray\x18\x04 \x03(\x0b\x32\x13.GeoReferencedArray*\xd3\x01\n\x08\x41xisType\x12\x0b\n\x07RunTime\x10\x00\x12\x0c\n\x08\x45nsemble\x10\x01\x12\x08\n\x04Time\x10\x02\x12\x08\n\x04GeoX\x10\x03\x12\x08\n\x04GeoY\x10\x04\x12\x08\n\x04GeoZ\x10\x05\x12\x07\n\x03Lat\x10\x06\x12\x07\n\x03Lon\x10\x07\x12\n\n\x06Height\x10\x08\x12\x0c\n\x08Pressure\x10\t\x12\x11\n\rRadialAzimuth\x10\n\x12\x12\n\x0eRadialDistance\x10\x0b\x12\x13\n\x0fRadialElevation\x10\x0c\x12\x0c\n\x08Spectral\x10\r\x12\x0e\n\nTimeOffset\x10\x0e*{\n\x0b\x41xisSpacing\x12\x10\n\x0cregularPoint\x10\x00\x12\x12\n\x0eirregularPoint\x10\x01\x12\x16\n\x12\x63ontiguousInterval\x10\x02\x12\x19\n\x15\x64iscontiguousInterval\x10\x03\x12\x13\n\x0fregularInterval\x10\x04*S\n\x0e\x44\x65pendenceType\x12\x0f\n\x0bindependent\x10\x00\x12\r\n\tdependent\x10\x01\x12\n\n\x06scalar\x10\x02\x12\x08\n\x04twoD\x10\x03\x12\x0b\n\x07\x66mrcReg\x10\x04*t\n\x08\x43\x61lendar\x12\x17\n\x13proleptic_gregorian\x10\x00\x12\r\n\tgregorian\x10\x01\x12\n\n\x06noleap\x10\x02\x12\x0c\n\x08\x61ll_leap\x10\x03\x12\x10\n\x0cuniform30day\x10\x04\x12\n\n\x06julian\x10\x05\x12\x08\n\x04none\x10\x06*K\n\x0c\x43overageType\x12\x0b\n\x07General\x10\x00\x12\x0f\n\x0b\x43urvilinear\x10\x01\x12\x08\n\x04Grid\x10\x02\x12\t\n\x05Swath\x10\x03\x12\x08\n\x04\x46mrc\x10\x04\x42\x30\n\x1cucar.nc2.ft2.coverage.remoteB\x10\x43\x64mrFeatureProtob\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cdmrfeature_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034ucar.nc2.ft2.coverage.remoteB\020CdmrFeatureProto'
  _AXISTYPE._serialized_start=1572
  _AXISTYPE._serialized_end=1783
  _AXISSPACING._serialized_start=1785
  _AXISSPACING._serialized_end=1908
  _DEPENDENCETYPE._serialized_start=1910
  _DEPENDENCETYPE._serialized_end=1993
  _CALENDAR._serialized_start=1995
  _CALENDAR._serialized_end=2111
  _COVERAGETYPE._serialized_start=2113
  _COVERAGETYPE._serialized_end=2188
  _COORDTRANSFORM._serialized_start=37
  _COORDTRANSFORM._serialized_end=112
  _COORDAXIS._serialized_start=115
  _COORDAXIS._serialized_end=449
  _COORDSYS._serialized_start=451
  _COORDSYS._serialized_end=555
  _CALENDARDATERANGE._serialized_start=557
  _CALENDARDATERANGE._serialized_end=633
  _RECTANGLE._serialized_start=635
  _RECTANGLE._serialized_end=706
  _COVERAGE._serialized_start=709
  _COVERAGE._serialized_end=842
  _COVERAGEDATASET._serialized_start=845
  _COVERAGEDATASET._serialized_end=1168
  _GEOREFERENCEDARRAY._serialized_start=1171
  _GEOREFERENCEDARRAY._serialized_end=1403
  _COVERAGEDATARESPONSE._serialized_start=1406
  _COVERAGEDATARESPONSE._serialized_end=1569
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2016 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Provide access to the CDMRemoteFeature endpoint on TDS."""

from io import BytesIO

from .ncstream import read_cdmrf_messages
from ..ncss import NCSS


class CDMRemoteFeature(NCSS):
    """Communicate to the CDMRemoteFeature HTTP endpoint."""

    @staticmethod
    def _parse_messages(resp):
        """Parse server responses as CDMRemoteFeature messages."""
        return read_cdmrf_messages(BytesIO(resp))

    def _get_metadata(self):
        """Get header information and store as metadata for the endpoint."""
        self.metadata = self.fetch_header()
        self.variables = {g.name for g in self.metadata.grids}

    def fetch_header(self):
        """Make a header request to the endpoint."""
        query = self.query().add_query_parameter(req='header')
        return self._parse_messages(self.get_query(query).content)[0]

    def fetch_feature_type(self):
        """Request the featureType from the endpoint."""
        query = self.query().add_query_parameter(req='featureType')
        return self.get_query(query).content

    def fetch_coords(self, query):
        """Pull down coordinate data from the endpoint."""
        q = query.add_query_parameter(req='coord')
        return self._parse_messages(self.get_query(q).content)

    def get_data(self, query):
        """Pull down data (coverages) from the endpoint."""
        return self._parse_messages(self.get_data_raw(query))

    def get_data_raw(self, query):
        """Pull down the data but don't parse."""
        return self.get_query(query.add_query_parameter(req='data')).content



# Copyright (c) 2013-2019 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Read upper air data from the IA State archives."""

from datetime import datetime
import json
import warnings

import numpy as np
import numpy.ma as ma
import pandas as pd

from .._tools import get_wind_components
from ..http_util import HTTPEndPoint


class IAStateUpperAir(HTTPEndPoint):
    """Download and parse data from the Iowa State's upper air archive."""

    def __init__(self):
        """Set up endpoint."""
        super().__init__('http://mesonet.agron.iastate.edu/json')

    @classmethod
    def request_data(cls, time, site_id, interp_nans=False, **kwargs):
        """Retrieve upper air observations from Iowa State's archive for a single station.

        Parameters
        ----------
        time : datetime.datetime
            The date and time of the desired observation.

        site_id : str
            The three letter ICAO identifier of the station for which data should be
            downloaded.

        interp_nans : bool
            Flag to interpolate temperature and dewpoint at significant wind levels,
            which will otherwise be NaNs. Default is False.

        kwargs
            Arbitrary keyword arguments to use to initialize source

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        endpoint = cls()
        df = endpoint._get_data(time, site_id, None, **kwargs)
        if interp_nans:
            df['temperature'] = df['temperature'].interpolate()
            df['dewpoint'] = df['dewpoint'].interpolate()

        return df

    @classmethod
    def request_all_data(cls, time, pressure=None, **kwargs):
        """Retrieve upper air observations from Iowa State's archive for all stations.

        Parameters
        ----------
        time : datetime.datetime
            The date and time of the desired observation.

        pressure : float, optional
            The mandatory pressure level at which to request data (in hPa). If none is given,
            all the available data in the profiles is returned.

        kwargs
            Arbitrary keyword arguments to use to initialize source

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        endpoint = cls()
        df = endpoint._get_data(time, None, pressure, **kwargs)
        return df

    def _get_data(self, time, site_id, pressure=None):
        """Download data from Iowa State's upper air archive.

        Parameters
        ----------
        time : datetime.datetime
            Date and time for which data should be downloaded
        site_id : str
            Site id for which data should be downloaded
        pressure : float, optional
            Mandatory pressure level at which to request data (in hPa).

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        json_data = self._get_data_raw(time, site_id, pressure)
        data = {}
        for profile in json_data['profiles']:
            for pt in profile['profile']:
                for field in ('drct', 'dwpc', 'hght', 'pres', 'sknt', 'tmpc'):
                    data.setdefault(field, []).append(np.nan if pt[field] is None
                                                      else pt[field])
                for field in ('station', 'valid'):
                    data.setdefault(field, []).append(np.nan if profile[field] is None
                                                      else profile[field])

        # Make sure that the first entry has a valid temperature and dewpoint
        idx = np.argmax(~(np.isnan(data['tmpc']) | np.isnan(data['dwpc'])))

        # Stuff data into a pandas dataframe
        df = pd.DataFrame()
        df['pressure'] = ma.masked_invalid(data['pres'][idx:])
        df['height'] = ma.masked_invalid(data['hght'][idx:])
        df['temperature'] = ma.masked_invalid(data['tmpc'][idx:])
        df['dewpoint'] = ma.masked_invalid(data['dwpc'][idx:])
        df['direction'] = ma.masked_invalid(data['drct'][idx:])
        df['speed'] = ma.masked_invalid(data['sknt'][idx:])
        df['station'] = data['station'][idx:]
        df['time'] = [datetime.strptime(valid, '%Y-%m-%dT%H:%M:%SZ')
                      for valid in data['valid'][idx:]]

        # Calculate the u and v winds
        df['u_wind'], df['v_wind'] = get_wind_components(df['speed'],
                                                         np.deg2rad(df['direction']))

        # Drop any rows with all NaN values for T, Td, winds
        df = df.dropna(subset=('temperature', 'dewpoint', 'direction', 'speed',
                               'u_wind', 'v_wind'), how='all').reset_index(drop=True)

        # Add unit dictionary
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = {'pressure': 'hPa',
                        'height': 'meter',
                        'temperature': 'degC',
                        'dewpoint': 'degC',
                        'direction': 'degrees',
                        'speed': 'knot',
                        'u_wind': 'knot',
                        'v_wind': 'knot',
                        'station': None,
                        'time': None}

        return df

    def _get_data_raw(self, time, site_id, pressure=None):
        r"""Download data from the Iowa State's upper air archive.

        Parameters
        ----------
        time : datetime.datetime
            Date and time for which data should be downloaded
        site_id : str
            Site id for which data should be downloaded
        pressure : float, optional
            Mandatory pressure level at which to request data (in hPa).

        Returns
        -------
        list of json data

        """
        query = {'ts': time.strftime('%Y%m%d%H00')}
        if site_id is not None:
            query['station'] = site_id
        if pressure is not None:
            query['pressure'] = pressure

        resp = self.get_path('raob.py', query)
        json_data = json.loads(resp.text)

        # See if the return is valid, but has no data
        if not (json_data['profiles'] and json_data['profiles'][0]['profile']):
            message = 'No data available '
            if time is not None:
                message += f'for {time: %Y-%m-%d %HZ} '
            if site_id is not None:
                message += f'for station {site_id}'
            if pressure is not None:
                message += f'for pressure {pressure}'
            message = message + '.'
            raise ValueError(message)
        return json_data



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Support making data requests to the radar data query service (radar server) on a TDS.

This includes forming proper queries as well as parsing the returned catalog.
"""

from collections import namedtuple
import xml.etree.ElementTree as ET  # noqa:N814

from .catalog import TDSCatalog
from .http_util import BadQueryError, DataQuery, HTTPEndPoint, urljoin


class RadarQuery(DataQuery):
    """Represent a query to the THREDDS radar server.

    Expands on the queries supported by :class:`~siphon.http_util.DataQuery` to add queries
    specific to the radar data query service.
    """

    def stations(self, *stns):
        """Specify one or more stations for the query.

        This modifies the query in-place, but returns `self` so that multiple
        queries can be chained together on one line.

        This replaces any existing spatial queries that have been set.

        Parameters
        ----------
        stns : str
            One or more names of variables to request

        Returns
        -------
        self : RadarQuery
            Returns self for chaining calls

        """
        self._set_query(self.spatial_query, stn=stns)
        return self


class RadarServer(HTTPEndPoint):
    """Wrap access to the THREDDS radar query service (radar server).

    Simplifies access via HTTP to the radar server endpoint. Parses the metadata, provides
    query catalog results download and parsing based on the appropriate query.

    Attributes
    ----------
    metadata : :class:`~siphon.metadata.TDSCatalogMetadata`
        Contains the result of parsing the radar server endpoint's dataset.xml. This has
        information about the time and space coverage, as well as full information
        about all of the variables.
    variables : set(str)
        Names of all variables available in this dataset
    stations : dict[str, Station]
        Mapping of station ID to a :class:`Station`, which is a namedtuple containing the
        station's id, name, latitude, longitude, and elevation.

    """

    def __init__(self, url):
        """Create a RadarServer instance.

        Parameters
        ----------
        url : str
            The base URL for the endpoint

        """
        xmlfile = '/dataset.xml'
        if url.endswith(xmlfile):
            url = url[:-len(xmlfile)]
        super().__init__(url)

    def _get_metadata(self):
        ds_cat = TDSCatalog(self.url_path('dataset.xml'))
        self.metadata = ds_cat.metadata
        self.variables = {k.split('/')[0] for k in self.metadata['variables']}
        self._get_stations()

    def _get_stations(self, station_file='stations.xml'):
        resp = self.get_path(station_file)
        self.stations = parse_station_table(ET.fromstring(resp.content))

    def query(self):
        """Return a new query for the radar server.

        Returns
        -------
        RadarQuery
            The new query

        """
        return RadarQuery()

    def validate_query(self, query):
        """Validate a query.

        Determines whether `query` is well-formed. This includes checking for all
        required parameters, as well as checking parameters for valid values.

        Parameters
        ----------
        query : RadarQuery
            The query to validate

        Returns
        -------
        valid : bool
            Whether `query` is valid.

        """
        valid = True
        # Make sure all stations are in the table
        if 'stn' in query.spatial_query:
            valid = valid and all(stid in self.stations
                                  for stid in query.spatial_query['stn'])

        if query.var:
            valid = valid and all(var in self.variables for var in query.var)

        return valid

    def get_catalog(self, query):
        """Fetch a parsed THREDDS catalog from the radar server.

        Requests a catalog of radar data files data from the radar server given the
        parameters in `query` and returns a :class:`~siphon.catalog.TDSCatalog` instance.

        Parameters
        ----------
        query : RadarQuery
            The parameters to send to the radar server

        Returns
        -------
        catalog : TDSCatalog
            The catalog of matching data files

        Raises
        ------
        :class:`~siphon.http_util.BadQueryError`
            When the query cannot be handled by the server

        See Also
        --------
        get_catalog_raw

        """
        # TODO: Refactor TDSCatalog so we don't need two requests, or to do URL munging
        try:
            url = self._base[:-1] if self._base[-1] == '/' else self._base
            url += '?' + str(query)
            return TDSCatalog(url)
        except ET.ParseError as e:
            raise BadQueryError(self.get_catalog_raw(query)) from e

    def get_catalog_raw(self, query):
        """Fetch THREDDS catalog XML from the radar server.

        Requests a catalog of radar data files data from the radar server given the
        parameters in `query` and returns the raw XML.

        Parameters
        ----------
        query : RadarQuery
            The parameters to send to the radar server

        Returns
        -------
        catalog : bytes
            The XML of the catalog of matching data files

        See Also
        --------
        get_catalog

        """
        return self.get_query(query).content


def get_radarserver_datasets(server):
    """Get datasets from a THREDDS radar server's top-level catalog.

    This is a helper function to construct the appropriate catalog URL from the
    server URL, fetch the catalog, and return the contained catalog references.

    Parameters
    ----------
    server : str
        The base URL to the THREDDS server

    Returns
    -------
    datasets : dict[str, :class:`~siphon.catalog.CatalogRef`]
        Mapping of dataset name to the catalog reference

    """
    if server[-1] != '/':
        server += '/'
    return TDSCatalog(urljoin(server, 'radarServer/catalog.xml')).catalog_refs


#
# The remainder of the file is not considered part of the public API.
# Use at your own risk!
#

Station = namedtuple('Station', 'id elevation latitude longitude name')


def parse_station_table(root):
    """Parse station list XML file."""
    stations = [parse_xml_station(elem) for elem in root.findall('station')]
    return {st.id: st for st in stations}


def parse_xml_station(elem):
    """Create a :class:`Station` instance from an XML tag."""
    stid = elem.attrib['id']
    name = elem.find('name').text
    lat = float(elem.find('latitude').text)
    lon = float(elem.find('longitude').text)
    elev = float(elem.find('elevation').text)
    return Station(id=stid, elevation=elev, latitude=lat, longitude=lon, name=name)



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Read data from the National Data Buoy Center."""

from io import StringIO
import warnings

import numpy as np
import pandas as pd
import requests

from ..http_util import HTTPEndPoint


class NDBC(HTTPEndPoint):
    """Download and parse data from the National Data Buoy Center."""

    def __init__(self):
        """Set up endpoint."""
        super().__init__('https://www.ndbc.noaa.gov/')

    @classmethod
    def realtime_observations(cls, buoy, data_type='txt'):
        """Retrieve the realtime buoy data from NDBC.

        Parameters
        ----------
        buoy : str
            Name of buoy
        data_type : str
            Type of data requested, must be one of
            'txt' standard meteorological data
            'drift' meteorological data from drifting buoys and limited moored buoy data
            mainly from international partners
            'cwind' continuous winds data (10 minute average)
            'spec' spectral wave summaries
            'ocean' oceanographic data
            'srad' solar radiation data
            'dart' water column height
            'supl' supplemental measurements data
            'rain' hourly rain data

        Returns
        -------
        `pandas.DataFrame`
            Parsed data

        """
        endpoint = cls()
        parsers = {'txt': endpoint._parse_met,
                   'drift': endpoint._parse_drift,
                   'cwind': endpoint._parse_cwind,
                   'spec': endpoint._parse_spec,
                   'ocean': endpoint._parse_ocean,
                   'srad': endpoint._parse_srad,
                   'dart': endpoint._parse_dart,
                   'supl': endpoint._parse_supl,
                   'rain': endpoint._parse_rain}

        if data_type not in parsers:
            raise KeyError('Data type must be txt, drift, cwind, spec, ocean, srad, dart,'
                           'supl, or rain for parsed realtime data.')

        raw_data = endpoint.raw_buoy_data(buoy, data_type=data_type)
        return parsers[data_type](raw_data)

    @staticmethod
    def _parse_met(content):
        """Parse standard meteorological data from NDBC buoys.

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour', 'minute',
                     'wind_direction', 'wind_speed', 'wind_gust',
                     'wave_height', 'dominant_wave_period', 'average_wave_period',
                     'dominant_wave_direction', 'pressure',
                     'air_temperature', 'water_temperature', 'dewpoint',
                     'visibility', '3hr_pressure_tendency', 'water_level_above_mean']

        col_units = {'wind_direction': 'degrees',
                     'wind_speed': 'meters/second',
                     'wind_gust': 'meters/second',
                     'wave_height': 'meters',
                     'dominant_wave_period': 'seconds',
                     'average_wave_period': 'seconds',
                     'dominant_wave_direction': 'degrees',
                     'pressure': 'hPa',
                     'air_temperature': 'degC',
                     'water_temperature': 'degC',
                     'dewpoint': 'degC',
                     'visibility': 'nautical_mile',
                     '3hr_pressure_tendency': 'hPa',
                     'water_level_above_mean': 'feet',
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')
        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _parse_drift(content):
        """Parse meteorological data from drifting buoys and limited moored buoy data.

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour_minute',
                     'latitude', 'longitude',
                     'wind_direction', 'wind_speed', 'wind_gust',
                     'pressure', '3hr_pressure_tendency',
                     'air_temperature', 'water_temperature']

        col_units = {'latitude': 'degrees',
                     'longitude': 'degrees',
                     'wind_direction': 'degrees',
                     'wind_speed': 'meters/second',
                     'wind_gust': 'meters/second',
                     'pressure': 'hPa',
                     'air_temperature': 'degC',
                     'water_temperature': 'degC',
                     '3hr_pressure_tendency': 'hPa',
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')
        df['hour'] = np.floor(df['hour_minute'] / 100)
        df['minute'] = df['hour_minute'] - df['hour'] * 100
        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour_minute', 'hour', 'minute'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _parse_cwind(content):
        """Parse continuous wind data (10 minute average).

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour', 'minute',
                     'wind_direction', 'wind_speed', 'gust_direction',
                     'wind_gust', 'gust_time']

        col_units = {'wind_direction': 'degrees',
                     'wind_speed': 'meters/second',
                     'gust_direction': 'degrees',
                     'wind_gust': 'meters/second',
                     'gust_time': None,
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')
        df['gust_direction'] = df['gust_direction'].replace(999, np.nan)
        df['wind_gust'] = df['wind_gust'].replace(99.0, np.nan)
        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)
        df['hours'] = np.floor(df['gust_time'] / 100)
        df['minutes'] = df['gust_time'] - df['hours'] * 100
        df['hours'] = df['hours'].replace(99, np.nan)
        df['minutes'] = df['minutes'].replace(99, np.nan)
        df['gust_time'] = pd.to_datetime(df[['year', 'month', 'day', 'hours', 'minutes']],
                                         utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute',
                              'hours', 'minutes'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _parse_spec(content):
        """Parse spectral wave summaries.

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour', 'minute',
                     'significant_wave_height', 'swell_height',
                     'swell_period', 'wind_wave_height', 'wind_wave_period',
                     'swell_direction', 'wind_wave_direction', 'steepness',
                     'average_wave_period', 'dominant_wave_direction']

        col_units = {'significant_wave_height': 'meters',
                     'swell_height': 'meters',
                     'swell_period': 'seconds',
                     'wind_wave_height': 'meters',
                     'wind_wave_period': 'seconds',
                     'swell_direction': None,
                     'wind_wave_direction': None,
                     'steepness': None,
                     'average_wave_period': 'seconds',
                     'dominant_wave_direction': 'degrees',
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')
        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _parse_ocean(content):
        """Parse oceanographic data.

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour', 'minute',
                     'measurement_depth', 'ocean_temperature',
                     'conductivity', 'salinity', 'oxygen_concentration',
                     'oxygen_concentration_ppm', 'chlorophyll_concentration',
                     'turbidity', 'pH', 'Eh']

        col_units = {'measurement_depth': 'meters',
                     'ocean_temperature': 'degC',
                     'conductivity': 'milliSiemens/centimeter',
                     'salinity': 'psu',
                     'oxygen_concentration': 'percent',
                     'oxygen_concentration_ppm': 'ppm',
                     'chlorophyll_concentration': 'micrograms/liter',
                     'turbidity': 'ftu',
                     'pH': 'dimensionless',
                     'Eh': 'millivolts',
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')
        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _parse_srad(content):
        """Parse solar radiation data.

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour', 'minute',
                     'shortwave_radiation_licor', 'shortwave_radiation_eppley',
                     'longwave_radiation']

        col_units = {'shortwave_radiation_licor': 'watts/meter^2',
                     'shortwave_radiation_eppley': 'watts/meter^2',
                     'longwave_radiation': 'watts/meter^2',
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')
        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _parse_dart(content):
        """Parse water column height data.

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour', 'minute', 'second',
                     'measurement_type', 'height']

        col_units = {'measurement_type': 'minutes',
                     'height': 'meters',
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')

        # Replace measurement type integer with minute value
        # 1 = 15-minute measurement
        # 2 = 1-minute measurement
        # 3 = 15-second measurement
        df['measurement_type'] = df['measurement_type'].replace(1, 15)
        df['measurement_type'] = df['measurement_type'].replace(2, 1)
        df['measurement_type'] = df['measurement_type'].replace(3, 0.25)

        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour',
                                        'minute', 'second']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute', 'second'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _parse_rain(content):
        """Parse hourly rain data.

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour', 'minute',
                     'hourly_accumulation']

        col_units = {'hourly_accumulation': 'millimeters',
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')

        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _parse_supl(content):
        """Parse supplemental measurements data.

        Parameters
        ----------
        content : str
            Data to parse

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        col_names = ['year', 'month', 'day', 'hour', 'minute',
                     'hourly_low_pressure', 'hourly_low_pressure_time',
                     'hourly_high_wind', 'hourly_high_wind_direction',
                     'hourly_high_wind_time']

        col_units = {'hourly_low_pressure': 'hPa',
                     'hourly_low_pressure_time': None,
                     'hourly_high_wind': 'meters/second',
                     'hourly_high_wind_direction': 'degrees',
                     'hourly_high_wind_time': None,
                     'time': None}

        df = pd.read_csv(StringIO(content), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')

        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)

        df['hours'] = np.floor(df['hourly_low_pressure_time'] / 100)
        df['minutes'] = df['hourly_low_pressure_time'] - df['hours'] * 100
        df['hours'] = df['hours'].replace(99, np.nan)
        df['minutes'] = df['minutes'].replace(99, np.nan)
        df['hourly_low_pressure_time'] = pd.to_datetime(df[['year', 'month', 'day', 'hours',
                                                            'minutes']], utc=True)

        df['hours'] = np.floor(df['hourly_high_wind_time'] / 100)
        df['minutes'] = df['hourly_high_wind_time'] - df['hours'] * 100
        df['hours'] = df['hours'].replace(99, np.nan)
        df['minutes'] = df['minutes'].replace(99, np.nan)
        df['hourly_high_wind_time'] = pd.to_datetime(df[['year', 'month', 'day',
                                                         'hours', 'minutes']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute', 'hours', 'minutes'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @classmethod
    def latest_observations(cls):
        """Retrieve the latest observations for all buoys from NDBC.

        Returns
        -------
            :class:`pandas.DataFrame` containing the data

        """
        endpoint = cls()
        col_names = ['station', 'latitude', 'longitude',
                     'year', 'month', 'day', 'hour', 'minute',
                     'wind_direction', 'wind_speed', 'wind_gust',
                     'wave_height', 'dominant_wave_period', 'average_wave_period',
                     'dominant_wave_direction', 'pressure', '3hr_pressure_tendency',
                     'air_temperature', 'water_temperature', 'dewpoint',
                     'visibility', 'water_level_above_mean']

        col_units = {'station': None,
                     'latitude': 'degrees',
                     'longitude': 'degrees',
                     'wind_direction': 'degrees',
                     'wind_speed': 'meters/second',
                     'wind_gust': 'meters/second',
                     'wave_height': 'meters',
                     'dominant_wave_period': 'seconds',
                     'average_wave_period': 'seconds',
                     'dominant_wave_direction': 'degrees',
                     'pressure': 'hPa',
                     'air_temperature': 'degC',
                     'water_temperature': 'degC',
                     'dewpoint': 'degC',
                     'visibility': 'nautical_mile',
                     '3hr_pressure_tendency': 'hPa',
                     'water_level_above_mean': 'feet',
                     'time': None}
        resp = endpoint.get_path('data/latest_obs/latest_obs.txt')

        df = pd.read_csv(StringIO(resp.text), comment='#', na_values='MM',
                         names=col_names, sep=r'\s+')
        df['time'] = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']], utc=True)
        df = df.drop(columns=['year', 'month', 'day', 'hour', 'minute'])
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', "Pandas doesn't allow columns to be created",
                                    UserWarning)
            df.units = col_units

        return df

    @staticmethod
    def _check_if_url_valid(url):
        """Check if a url is valid (returns 200) or not.

        Parameters
        ----------
        url : str
            URL to check

        Returns
        -------
            bool if url is valid

        """
        r = requests.head(url, timeout=300)
        return r.status_code == 200

    @classmethod
    def buoy_data_types(cls, buoy):
        """Determine which types of data are available for a given buoy.

        Parameters
        ----------
        buoy : str
            Buoy name

        Returns
        -------
        dict[str, str]
            dict of valid file extensions and their descriptions

        """
        endpoint = cls()
        file_types = {'txt': 'standard meteorological data',
                      'drift': 'meteorological data from drifting buoys and limited moored'
                               'buoy data mainly from international partners',
                      'cwind': 'continuous wind data (10 minute average)',
                      'spec': 'spectral wave summaries',
                      'data_spec': 'raw spectral wave data',
                      'swdir': 'spectral wave data (alpha1)',
                      'swdir2': 'spectral wave data (alpha2)',
                      'swr1': 'spectral wave data (r1)',
                      'swr2': 'spectral wave data (r2)',
                      'adcp': 'acoustic doppler current profiler',
                      'ocean': 'oceanographic data',
                      'tide': 'tide data',
                      'srad': 'solar radiation data',
                      'dart': 'water column height',
                      'supl': 'supplemental measurements data',
                      'rain': 'hourly rain data'}
        available_data = {}
        buoy_url = 'https://www.ndbc.noaa.gov/data/realtime2/' + buoy + '.'
        for key in file_types:
            if endpoint._check_if_url_valid(buoy_url + key):
                available_data[key] = file_types[key]
        return available_data

    @classmethod
    def raw_buoy_data(cls, buoy, data_type='txt'):
        """Retrieve the raw buoy data contents from NDBC.

        Parameters
        ----------
        buoy : str
            Name of buoy

        data_type : str
            Type of data requested, must be one of
            'txt' standard meteorological data
            'drift' meteorological data from drifting buoys and limited moored buoy data
            mainly from international partners
            'cwind' continuous winds data (10 minute average)
            'spec' spectral wave summaries
            'data_spec' raw spectral wave data
            'swdir' spectral wave data (alpha1)
            'swdir2' spectral wave data (alpha2)
            'swr1' spectral wave data (r1)
            'swr2' spectral wave data (r2)
            'adcp' acoustic doppler current profiler
            'ocean' oceanographic data
            'tide' tide data
            'srad' solar radiation data
            'dart' water column height
            'supl' supplemental measurements data
            'rain' hourly rain data

        Returns
        -------
        str
            Raw data string

        """
        endpoint = cls()
        resp = endpoint.get_path(f'data/realtime2/{buoy}.{data_type}')
        return resp.text



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Support making data requests to the NetCDF subset service (NCSS) on a TDS.

This includes forming proper queries as well as parsing the returned data.
"""

import atexit
from io import BytesIO
from os import remove
import platform
import xml.etree.ElementTree as ET  # noqa:N814

import numpy as np

from .http_util import DataQuery, HTTPEndPoint, parse_iso_date
from .ncss_dataset import NCSSDataset


def default_unit_handler(data, units=None):  # pylint:disable=unused-argument
    """Handle units in the default manner.

    Ignores units and just returns :func:`numpy.array`.
    """
    return np.array(data)


class NCSS(HTTPEndPoint):
    """Wrap access to the NetCDF Subset Service (NCSS) on a THREDDS server.

    Simplifies access via HTTP to the NCSS endpoint. Parses the metadata, provides
    data download and parsing based on the appropriate query.

    Attributes
    ----------
    metadata : NCSSDataset
        Contains the result of parsing the NCSS endpoint's dataset.xml. This has
        information about the time and space coverage, as well as full information
        about all of the variables.
    variables : set(str)
        Names of all variables available in this dataset
    unit_handler : callable
        Function to handle units that come with CSV/XML data. Should be a callable that
        takes a list of string values and unit str (can be :data:`None`), and returns the
        desired representation of values. Defaults to ignoring units and returning
        :func:`numpy.array`.

    """

    # Need staticmethod to keep this from becoming a bound method, where self
    # is passed implicitly. Needed to avoid warning about duplicated docstring.
    unit_handler = staticmethod(lambda *a, **kw: default_unit_handler(*a, **kw))

    def _get_metadata(self):
        # Need to use .content here to avoid decode problems
        meta_xml = self.get_path('dataset.xml').content
        root = ET.fromstring(meta_xml)
        self.metadata = NCSSDataset(root)
        self.variables = set(self.metadata.variables)

    def query(self):
        """Return a new query for NCSS.

        Returns
        -------
        query : NCSSQuery
            The newly created query

        """
        return NCSSQuery()

    def validate_query(self, query):
        """Validate a query.

        Determines whether `query` is well-formed. This includes checking for all
        required parameters, as well as checking parameters for valid values.

        Parameters
        ----------
        query : NCSSQuery
            The query to validate

        Returns
        -------
        valid : bool
            Whether `query` is valid.

        """
        # Make sure all variables are in the dataset
        return bool(query.var) and all(var in self.variables for var in query.var)

    def get_data(self, query):
        """Fetch parsed data from a THREDDS server using NCSS.

        Requests data from the NCSS endpoint given the parameters in `query` and
        handles parsing of the returned content based on the mimetype.

        Parameters
        ----------
        query : NCSSQuery
            The parameters to send to the NCSS endpoint

        Returns
        -------
        Parsed data response from the server. Exact format depends on the format of the
        response.

        See Also
        --------
        get_data_raw

        """
        resp = self.get_query(query)
        return response_handlers(resp, self.unit_handler)

    def get_data_raw(self, query):
        """Fetch raw data from a THREDDS server using NCSS.

        Requests data from the NCSS endpoint given the parameters in `query` and
        returns the raw bytes of the response.

        Parameters
        ----------
        query : NCSSQuery
            The parameters to send to the NCSS endpoint

        Returns
        -------
        content : bytes
            The raw, un-parsed, data returned by the server

        See Also
        --------
        get_data

        """
        return self.get_query(query).content


class NCSSQuery(DataQuery):
    """Represent a query to the NetCDF Subset Service (NCSS).

    Expands on the queries supported by :class:`~siphon.http_util.DataQuery` to add queries
    specific to NCSS.
    """

    def projection_box(self, min_x, min_y, max_x, max_y):
        """Add a bounding box in projected (native) coordinates to the query.

        This adds a request for a spatial bounding box, bounded by (`min_x`, `max_x`) for
        x direction and (`min_y`, `max_y`) for the y direction. This modifies the query
        in-place, but returns ``self`` so that multiple queries can be chained together
        on one line.

        This replaces any existing spatial queries that have been set.

        Parameters
        ----------
        min_x : float
            The left edge of the bounding box
        min_y : float
            The bottom edge of the bounding box
        max_x : float
            The right edge of the bounding box
        max_y: float
            The top edge of the bounding box

        Returns
        -------
        self : NCSSQuery
            Returns self for chaining calls

        """
        self._set_query(self.spatial_query, minx=min_x, miny=min_y,
                        maxx=max_x, maxy=max_y)
        return self

    def accept(self, fmt):
        """Set format for data returned from NCSS.

        This modifies the query in-place, but returns `self` so that multiple queries
        can be chained together on one line.

        Parameters
        ----------
        fmt : str
            The format to send to the server.

        Returns
        -------
        self : NCSSQuery
            Returns self for chaining calls

        """
        return self.add_query_parameter(accept=fmt)

    def add_lonlat(self, value=True):
        """Set whether NCSS should add latitude/longitude to returned data.

        This is only used on grid requests. Used to make returned data CF-compliant.
        This modifies the query in-place, but returns `self` so that multiple queries
        can be chained together on one line.

        Parameters
        ----------
        value : bool, optional
            Whether to add latitude/longitude information. Defaults to True.

        Returns
        -------
        self : NCSSQuery
            Returns self for chaining calls

        """
        return self.add_query_parameter(addLatLon=value)

    def strides(self, time=None, spatial=None):
        """Set time and/or spatial (horizontal) strides.

        This is only used on grid requests. Used to skip points in the returned data.
        This modifies the query in-place, but returns `self` so that multiple queries
        can be chained together on one line.

        Parameters
        ----------
        time : int, optional
            Stride for times returned. Defaults to None, which is equivalent to 1.
        spatial : int, optional
            Stride for horizontal grid. Defaults to None, which is equivalent to 1.

        Returns
        -------
        self : NCSSQuery
            Returns self for chaining calls

        """
        if time:
            self.add_query_parameter(timeStride=time)
        if spatial:
            self.add_query_parameter(horizStride=spatial)
        return self

    def vertical_level(self, level):
        """Set vertical level for which data should be retrieved.

        The value depends on the coordinate values for the vertical dimension of the
        requested variable.

        This modifies the query in-place, but returns `self` so that multiple queries
        can be chained together on one line.

        Parameters
        ----------
        level : float
            The value of the desired level

        Returns
        -------
        self : NCSSQuery
            Returns self for chaining calls

        """
        return self.add_query_parameter(vertCoord=level)


#
# The remainder of the file is not considered part of the public API.
# Use at your own risk!
#

class ResponseRegistry:
    """Register functions to be called based on the mimetype in the response headers."""

    def __init__(self):
        """Initialize the registry."""
        self._reg = {}

    def register(self, mimetype):
        """Register a function to handle a particular mimetype."""
        def dec(func):
            self._reg[mimetype] = func
            return func
        return dec

    @staticmethod
    def default(content, units):  # pylint:disable=unused-argument
        """Handle a mimetype when no function is registered."""
        return content

    def __call__(self, resp, unit_handler):
        """Process the HTTP response using the appropriate handler."""
        mimetype = resp.headers['content-type'].split(';')[0]
        return self._reg.get(mimetype, self.default)(resp.content, unit_handler)


response_handlers = ResponseRegistry()


def squish(seq):
    """If list contains only 1 element, return it instead."""
    return seq if len(seq) > 1 else seq[0]


def combine_dicts(seq):
    """Combine a list of dictionaries into single one."""
    ret = {}
    for item in seq:
        ret.update(item)
    return ret


# Parsing of XML returns from NCSS
@response_handlers.register('application/xml')
def parse_xml(data, handle_units):
    """Parse XML data returned by NCSS."""
    root = ET.fromstring(data)
    return squish(parse_xml_dataset(root, handle_units))


def parse_xml_point(elem):
    """Parse an XML point tag."""
    point = {}
    units = {}
    for data in elem.findall('data'):
        name = data.get('name')
        unit = data.get('units')
        point[name] = float(data.text) if name != 'date' else parse_iso_date(data.text)
        if unit:
            units[name] = unit
    return point, units


def combine_xml_points(seq, units, handle_units):
    """Combine multiple Point tags into an array."""
    ret = {}
    for item in seq:
        for key, value in item.items():
            ret.setdefault(key, []).append(value)

    for key, value in ret.items():
        if key != 'date':
            ret[key] = handle_units(value, units.get(key, None))

    return ret


def parse_xml_dataset(elem, handle_units):
    """Create a netCDF-like dataset from XML data."""
    points, units = zip(*[parse_xml_point(p) for p in elem.findall('point')], strict=False)
    # Group points by the contents of each point
    datasets = {}
    for p in points:
        datasets.setdefault(tuple(p), []).append(p)

    all_units = combine_dicts(units)
    return [combine_xml_points(d, all_units, handle_units) for d in datasets.values()]


# Handling of netCDF 3/4 from NCSS
try:
    from tempfile import NamedTemporaryFile

    from netCDF4 import Dataset

    @response_handlers.register('application/x-netcdf')
    @response_handlers.register('application/x-netcdf4')
    def read_netcdf(data, handle_units):  # pylint:disable=unused-argument
        """Handle HTTP responses in netCDF format."""
        ostype = platform.architecture()
        if ostype[1].lower() == 'windowspe':
            with NamedTemporaryFile(delete=False) as tmp_file:
                tmp_file.write(data)
                tmp_file.flush()
                atexit.register(deletetempfile, tmp_file.name)
                return Dataset(tmp_file.name, 'r')
        else:
            with NamedTemporaryFile() as tmp_file:
                tmp_file.write(data)
                tmp_file.flush()
                return Dataset(tmp_file.name, 'r')
except ImportError:
    import warnings
    warnings.warn('netCDF4 module not installed. '
                  'Will be unable to handle NetCDF returns from NCSS.', stacklevel=2)


def deletetempfile(fname):
    """Delete a temporary file.

    Warn on any exceptions.
    """
    try:
        remove(fname)
    except OSError:
        import warnings
        warnings.warn('temporary netcdf dataset file not deleted. '
                      'to delete temporary dataset file in the future '
                      'be sure to use dataset.close() when finished.', stacklevel=2)


# Parsing of CSV data returned from NCSS
@response_handlers.register('text/plain')
def parse_csv_response(data, unit_handler):
    """Handle CSV-formatted HTTP responses."""
    return squish([parse_csv_dataset(d, unit_handler) for d in data.split(b'\n\n')])


def parse_csv_header(line):
    """Parse the CSV header returned by TDS."""
    units = {}
    names = []
    for var in line.split(','):
        start = var.find('[')
        if start < 0:
            names.append(str(var))
            continue
        else:
            names.append(str(var[:start]))
        end = var.find(']', start)
        unitstr = var[start + 1:end]
        eq = unitstr.find('=')
        if eq >= 0:
            # go past = and ", skip final "
            units[names[-1]] = unitstr[eq + 2:-1]
    return names, units


# Only needed until we support only numpy >= 2.0.
def _decode_if_needed(s):
    try:
        return s.decode('utf-8')
    except AttributeError:
        return s


def parse_csv_dataset(data, handle_units):
    """Parse CSV data into a netCDF-like dataset."""
    fobj = BytesIO(data)
    names, units = parse_csv_header(fobj.readline().decode('utf-8'))
    arrs = np.genfromtxt(fobj, dtype=None, names=names, delimiter=',',
                         converters={'date': lambda s: parse_iso_date(_decode_if_needed(s))})
    d = {}
    for f in arrs.dtype.fields:
        dat = arrs[f]
        if dat.dtype == object:
            dat = dat.tolist()
        d[f] = handle_units(dat, units.get(f, None))
    return d



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Provide a netCDF4-like interface on top of CDMRemote and NCStream."""

from collections import OrderedDict
import enum
import logging

from .cdmremote import CDMRemote
from .ncstream import unpack_attribute, unpack_variable

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)


class AttributeContainer:
    """Unpack and provide access to attributes."""

    def __init__(self):
        """Initialize the container."""
        self._attrs = []

    def ncattrs(self):
        """Return a list of all available attributes."""
        return self._attrs

    def _unpack_attrs(self, attrs):
        for att in attrs:
            name, val = unpack_attribute(att)
            self._attrs.append(name)
            setattr(self, name, val)


class Group(AttributeContainer):
    """Group together variables, attributes, and dimensions."""

    def __init__(self, parent=None):
        """Initialize a Group."""
        super().__init__()
        self.groups = OrderedDict()
        self.variables = OrderedDict()
        self.dimensions = OrderedDict()
        self.types = OrderedDict()
        if parent:
            self.dataset = parent.dataset
        else:
            self.dataset = self

    @property
    def path(self):
        """Return the full path to the Group, including any parent Groups."""
        # If root, return '/'
        if self.dataset is self:
            return ''
        else:  # Otherwise recurse
            return self.dataset.path + '/' + self.name

    def load_from_stream(self, group):
        """Load a Group from an NCStream object."""
        self._unpack_attrs(group.atts)
        self.name = group.name

        for dim in group.dims:
            new_dim = Dimension(self, dim.name)
            self.dimensions[dim.name] = new_dim
            new_dim.load_from_stream(dim)

        for var in group.vars:
            new_var = Variable(self, var.name)
            self.variables[var.name] = new_var
            new_var.load_from_stream(var)

        for grp in group.groups:
            new_group = Group(self)
            self.groups[grp.name] = new_group
            new_group.load_from_stream(grp)

        for struct in group.structs:
            new_var = Variable(self, struct.name)
            self.variables[struct.name] = new_var
            new_var.load_from_stream(struct)

        if group.enumTypes:
            for en in group.enumTypes:
                self.types[en.name] = enum.Enum(en.name,
                                                [(typ.value, typ.code) for typ in en.map])

    def __str__(self):
        """Return a string representation of the Group and its members."""
        print_groups = []
        if self.name:
            print_groups.append(self.name)

        if self.groups:
            print_groups.append('Groups:')
            for group in self.groups.values():
                print_groups.append(str(group))
                print_groups.append('---end group---')

        if self.dimensions:
            print_groups.append('Dimensions:')
            for dim in self.dimensions.values():
                print_groups.append(str(dim))

        if self.types:
            print_groups.append('Types:')
            for name, typ in self.types.items():
                print_groups.append(name + ' ' + str(list(typ)))

        if self.variables:
            print_groups.append('Variables:')
            for var in self.variables.values():
                print_groups.append(str(var))

        if self.ncattrs():
            print_groups.append('Attributes:')
            for att in self.ncattrs():
                print_groups.append(f'\t{att}: {getattr(self, att)}')
        return '\n'.join(print_groups)

    __repr__ = __str__


class Dataset(Group):
    """Abstract away access to the remote dataset."""

    def __init__(self, url):
        """Initialize the dataset."""
        super().__init__()
        self.cdmr = CDMRemote(url)
        self.url = url
        self._read_header()

    def _read_header(self):
        messages = self.cdmr.fetch_header()
        if len(messages) != 1:
            log.warning('Receive %d messages for header!', len(messages))
        self._header = messages[0]
        self.load_from_stream(self._header.root)

    def __str__(self):
        """Return a string representation of the Dataset and all contained members."""
        return self.url + '\n' + super().__str__()

    __repr__ = __str__


class Variable(AttributeContainer):
    """Hold information about a data variable and provide access to the underlying data."""

    def __init__(self, group, name):
        """Initialize the Variable."""
        super().__init__()
        self._group = group
        self.name = name
        self.dimensions = ()
        self._data = None
        self.dataset = group.dataset
        self._enum = False

    def group(self):
        """Return the parent Group."""
        return self._group

    @property
    def path(self):
        """Return the full path to the Variable, including any parent Groups."""
        return self._group.path + '/' + self.name

    def __getitem__(self, ind):
        """Access the Variable's underlying data."""
        if self._data is not None:
            # For scalars, don't slice
            return self._data if not self.shape else self._data[ind]
        else:
            ind, keep_dims = self._process_indices(ind)

            # Get the data for our request. We assume we only get 1 message.
            messages = self.dataset.cdmr.fetch_data(**{self.path: ind})
            arr = messages[0]

            # Get the proper byte ordering.
            # We handle structures by looking for a structured dtype. By convention,
            # this has a single field which is has a void type and the byte order encoded
            # in its name. This is because we can't retrieve a useful byte order from
            # any of these flexible types.
            if arr.dtype == 'O' and hasattr(arr[0], 'dtype'):
                byteorder = arr[0].dtype.byteorder
            elif arr.dtype.fields and arr.dtype.names[0] in ('>', '<'):
                byteorder = arr.dtype.names[0]
            else:
                byteorder = arr.dtype.byteorder

            # Set the dtype on the returned data to our own dtype, with byte ordering set
            # based on what was returned. This allows us to handle structures.
            dt = self.dtype.newbyteorder(byteorder)

            if arr.dtype == 'O':
                if hasattr(arr[0], 'dtype') and arr[0].ndim > 0:
                    for i, subarray in enumerate(arr):
                        arr[i] = subarray.astype(dt)
                else:
                    arr = arr.astype(dt)
            # Don't reset dtype if we've already decoded to struct
            elif arr.dtype.fields and arr.dtype.fields == dt.fields:
                pass
            else:
                arr.dtype = dt

            # Need to handle removing dimensions that have had an index
            # applied -- the protocol returns them with size 1, but numpy
            # behavior removes them
            if keep_dims:
                return arr.reshape(*[arr.shape[i] for i in keep_dims])
            else:
                return arr.squeeze()

    def _process_indices(self, ind):
        # Make sure we have a list of indices
        try:
            ind = list(ind)
        except TypeError:
            ind = [ind]

        # Make sure we don't have too many things to index, but allow a full slice on
        # a scalar variable
        if len(ind) > self.ndim and (self.ndim != 0 or len(ind) != 1 or ind[0] != slice(None)):
                raise IndexError('Too many dimensions to index.')

        # Expand to a slice/ellipsis for every dimension
        if Ellipsis not in ind and len(ind) < self.ndim:
            ind.append(Ellipsis)

        # Check for ellipsis
        if Ellipsis in ind:
            num_empty = self.ndim - len(ind) + 1
            ellip_ind = ind.index(Ellipsis)
            ind.pop(ellip_ind)
            ind[ellip_ind:ellip_ind] = [slice(None)] * num_empty

        # Assemble the slices/indices
        keep_dims = []
        for dim, i in enumerate(ind):
            if isinstance(i, slice):
                is_vlen = self.dimensions and self.dimensions[dim] == '*'

                # Need to keep the dimension if we slice for non-vlen
                if not is_vlen:
                    keep_dims.append(dim)

                # Skip full slice
                if i.start is None and i.stop is None and i.step is None:
                    continue

                # vlen needs to have full slice
                if is_vlen:
                    raise RuntimeError("Can't slice along vlen dimension (%d)!", dim)

                # Adjust start and stop to handle negative indexing
                # and partial support for slicing beyond end.
                start = 0 if i.start is None else self._adjust_index(dim, i.start)

                stop = self.shape[dim] if i.stop is None else self._adjust_index(dim, i.stop)

                # Need to create new slice for adjusted values
                ind[dim] = slice(start, stop, i.step)
            else:
                # Adjust start and stop to handle negative indexing
                ind[dim] = self._adjust_index(dim, i)

        return ind, keep_dims

    def _adjust_index(self, dim, index):
        if index < 0:
            return self.shape[dim] + index
        elif index > self.shape[dim]:
            return self.shape[dim]
        return index

    def load_from_stream(self, var):
        """Populate the Variable from an NCStream object."""
        dims = []
        for d in var.shape:
            dim = Dimension(None, d.name)
            dim.load_from_stream(d)
            dims.append(dim)

        self.dimensions = tuple(dim.name for dim in dims)
        self.shape = tuple(dim.size for dim in dims)
        self.ndim = len(var.shape)
        self._unpack_attrs(var.atts)

        data, dt, type_name = unpack_variable(var)
        if data is not None:
            data = data.reshape(self.shape)
        self._data = data
        self.dtype = dt
        self.datatype = type_name

        if hasattr(var, 'enumType') and var.enumType:
            self.datatype = var.enumType
            self._enum = True

    def __str__(self):
        """Return a string representation of the Variable."""
        groups = [str(type(self))]
        groups.append('{} {}({})'.format(self.datatype, self.name,
                                         ', '.join(self.dimensions)))
        for att in self.ncattrs():
            groups.append(f'\t{att}: {getattr(self, att)}')
        if self.ndim:
            if self.ndim > 1:
                shape_str = '(' + ', '.join(str(s) for s in self.shape) + ')'
            else:
                shape_str = str(self.shape[0])
            groups.append('shape = ' + shape_str)
        return '\n'.join(groups)


class Dimension:
    """Hold information about dimensions shared between variables."""

    def __init__(self, group, name, size=None):
        """Initialize the Dimension."""
        self._group = group
        self.name = name

        self.size = size
        self.unlimited = False
        self.private = False
        self.vlen = False

    def group(self):
        """Return the parent Group."""
        return self._group

    def isunlimited(self):
        """Return whether the dimension is unlimited."""
        return self.unlimited

    def load_from_stream(self, dim):
        """Load from an NCStream object."""
        self.unlimited = dim.isUnlimited
        self.private = dim.isPrivate
        self.vlen = dim.isVlen
        if not self.vlen:
            self.size = dim.length

    def __len__(self):
        """Return the length of the Dimension."""
        return self.size if self.size is not None else 0

    def __str__(self):
        """Return a string representation of the Dimension information."""
        grps = [f'{type(self)} ']
        if self.unlimited:
            grps.append('(unlimited): ')

        if self.private:
            grps.append('(private)')
        else:
            grps.append('name = ' + self.name)

        if self.vlen:
            grps.append(', (vlen)')
        else:
            grps.append(f', size = {self.size}')

        return ''.join(grps)

    __repr__ = __str__



# Copyright (c) 2018 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Requests data from the ACIS Web Services API."""

import requests

from ..http_util import session_manager


def acis_request(method, params):
    """Request data from the ACIS Web Services API.

    Makes a request from the ACIS Web Services API for data
    based on a given method (StnMeta,StnData,MultiStnData,GridData,General)
    and parameters string. Information about the parameters can be obtained at:
    http://www.rcc-acis.org/docs_webservices.html

    If a connection to the API fails, then it will raise an exception. Some bad
    calls will also return empty dictionaries.

    ACIS Web Services is a distributed system! A call to the main URL can be
    delivered to any climate center running a public instance of the service.
    This makes the calls efficient, but also occasionally results in failed
    calls when a server you are directed to is having problems. Generally,
    reconnecting after waiting a few seconds will resolve a problem. If problems
    are persistent, contact ACIS developers at the High Plains Regional Climate
    Center or Northeast Regional Climate Center who will look into server
    issues.

    Parameters
    ----------
    method : str
        The Web Services request method (StnMeta, StnData, MultiStnData, GridData, General)
    params : dict
        A JSON array of parameters (See Web Services API)

    Returns
    -------
    dict[str, Any]
        A dictionary of data based on the JSON parameters

    Raises
    ------
    `AcisApiException`
        When the API is unable to establish a connection or returns
        unparsable data.

    """
    base_url = 'http://data.rcc-acis.org/'  # ACIS Web API URL

    timeout = 300 if method == 'MultiStnData' else 60

    try:
        response = session_manager.create_session().post(base_url + method, json=params,
                                                         timeout=timeout)
        return response.json()
    except requests.exceptions.Timeout as e:
        raise AcisApiException('Connection Timeout') from e
    except requests.exceptions.TooManyRedirects as e:
        raise AcisApiException('Bad URL. Check your ACIS connection method string.') from e
    except ValueError as e:
        raise AcisApiException('No data returned! The ACIS parameter dictionary'
                               'may be incorrectly formatted') from e


class AcisApiException(Exception):  # noqa: N818
    """Handle exceptions raised by the acis_request function."""




# Copyright (c) 2013-2019 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""
Code to support reading and parsing catalog files from a THREDDS Data Server (TDS).

They help identifying the latest dataset and finding proper URLs to access the data.
"""

from collections import OrderedDict
from datetime import datetime
import logging
import re
from urllib.parse import urljoin, urlparse
import warnings
import xml.etree.ElementTree as ET  # noqa:N814

from .http_util import session_manager
from .metadata import TDSCatalogMetadata

log = logging.getLogger(__name__)
log.setLevel(logging.ERROR)


class IndexableMapping(OrderedDict):
    """Extend ``OrderedDict`` to allow index-based access to values."""

    def __getitem__(self, item):
        """Return an item either by index or name."""
        try:
            item + ''  # Raises if item not a string
            return super().__getitem__(item)
        except TypeError:
            return list(self.values())[item]


class DatasetCollection(IndexableMapping):
    r"""Extend ``IndexableMapping`` to allow datetime-based filter queries.

    Indexing works like a dictionary. The dataset name ('my_data.nc', a string) is the key,
    and the value returned is an instance of ``Dataset``. Positional indexing
    (e.g., [0]) is another valid method of indexing.

    ``DatasetCollection`` is commonly encountered as the ``datasets`` attribute of a
    ``TDSCatalog``. If a ``regex`` in ``filter_time_nearest`` or ```filter_time_range` does not
    provide sufficient flexibility, or the ``TDSCatalog`` does not provide accurate times,
    iterating over ``datasets`` can be useful as part implementing a custom filter. For
    example, in ``for ds in catalog.datasets: print(ds)``, ``ds`` will be the dataset name, and
    ``ds`` can be used to implement further filtering logic.
    """

    default_regex = re.compile(r'(?P<year>\d{4})(?P<month>[01]\d)(?P<day>[0123]\d)_'
                               r'(?P<hour>[012]\d)(?P<minute>[0-5]\d)')

    def _get_datasets_with_times(self, regex, strptime=None):
        # Set the default regex if we don't have one
        # If strptime is provided, pass the regex group named 'strptime' to strptime
        regex = self.default_regex if regex is None else re.compile(regex)

        # Loop over the collection looking for keys that match our regex
        found_date = False
        for ds in self:
            match = regex.search(ds)

            # If we find one, make a datetime and yield it along with the value
            if match:
                found_date = True
                date_parts = match.groupdict()
                if strptime is not None:
                    date_str = date_parts.get('strptime', 0)
                    dt = datetime.strptime(date_str, strptime)
                else:
                    dt = datetime(int(date_parts.get('year', 0)),
                                  int(date_parts.get('month', 0)),
                                  int(date_parts.get('day', 0)),
                                  int(date_parts.get('hour', 0)),
                                  int(date_parts.get('minute', 0)),
                                  int(date_parts.get('second', 0)),
                                  int(date_parts.get('microsecond', 0)))
                yield dt, self[ds]

        # If we never found any keys that match, we should let the user know that rather
        # than have it be the same as if nothing matched filters
        if not found_date:
            raise ValueError('No datasets with times found.')

    def filter_time_nearest(self, time, regex=None, strptime=None):
        r"""Filter keys for an item closest to the desired time.

        Loops over all keys in the collection and uses `regex` to extract and build
        `datetime`s. The collection of `datetime`s is compared to `start` and the value that
        has a `datetime` closest to that requested is returned.If none of the keys in the
        collection match the regex, indicating that the keys are not date/time-based,
        a ``ValueError`` is raised.

        Parameters
        ----------
        time : ``datetime.datetime``
            The desired time
        regex : str, optional
            The regular expression to use to extract date/time information from the key. If
            given, this should contain either
            1. named groups: 'year', 'month', 'day', 'hour', 'minute', 'second',
            and 'microsecond', as appropriate. When a match is found, any of those groups
            missing from the pattern will be assigned a value of 0. The default pattern looks
            for patterns like: 20171118_2356.
            or
            2. a group named 'strptime' (e.g., r'_s(?P<strptime>\d{13})' for GOES-16 data)
            to be parsed with strptime.
        strptime : str, optional
            the format string that corresponds to regex option (2) above. For example, GOES-16
            data with a julian date matching the regex above is parsed with '%Y%j%H%M%S'.

        Returns
        -------
        Dataset
            The value with a time closest to that desired

        """
        return min(self._get_datasets_with_times(regex, strptime),
                   key=lambda i: abs((i[0] - time).total_seconds()))[-1]

    def filter_time_range(self, start, end, regex=None, strptime=None):
        r"""Filter keys for all items within the desired time range.

        Loops over all keys in the collection and uses `regex` to extract and build
        `datetime`s. From the collection of `datetime`s, all values within `start` and `end`
        (inclusive) are returned. If none of the keys in the collection match the regex,
        indicating that the keys are not date/time-based, a ``ValueError`` is raised.

        Parameters
        ----------
        start : ``datetime.datetime``
            The start of the desired time range, inclusive
        end : ``datetime.datetime``
            The end of the desired time range, inclusive
        regex : str, optional
            The regular expression to use to extract date/time information from the key. If
            given, this should contain either
            1. named groups: 'year', 'month', 'day', 'hour', 'minute', 'second',
            and 'microsecond', as appropriate. When a match is found, any of those groups
            missing from the pattern will be assigned a value of 0. The default pattern looks
            for patterns like: 20171118_2356.
            or
            2. a group named 'strptime' (e.g., r'_s(?P<strptime>\d{13})' for GOES-16 data)
            to be parsed with strptime.
        strptime : str, optional
            the format string that corresponds to regex option (2) above. For example, GOES-16
            data with a julian date matching the regex above is parsed with '%Y%j%H%M%S'.

        Returns
        -------
        List[Dataset]
            All values corresponding to times within the specified range

        """
        if start > end:
            warnings.warn('The provided start time comes after the end time. No data will '
                          'be returned.', UserWarning, stacklevel=2)
        return [item[-1] for item in self._get_datasets_with_times(regex, strptime)
                if start <= item[0] <= end]

    def __str__(self):
        """Return a string representation of the collection."""
        return str(list(self))

    __repr__ = __str__


def _try_lower(arg):
    try:
        arg = arg.lower()
    except (TypeError, AttributeError, ValueError):
        log.warning('Could not convert %s to lowercase.', arg)
    return arg


class CaseInsensitiveStr(str):
    """Extend ``str`` to use case-insensitive comparison and lookup."""

    def __init__(self, *args):
        """Create str with a _lowered property."""
        self._lowered = _try_lower(self)

    def __hash__(self):
        """Hash str using _lowered property."""
        return str.__hash__(self._lowered)

    def __eq__(self, other):
        """Return true if other is case-insensitive equal to self."""
        return str.__eq__(self._lowered, _try_lower(other))

    def __gt__(self, other):
        """Return true if other is case-insensitive greater than self."""
        return str.__gt__(self._lowered, _try_lower(other))

    def __ge__(self, other):
        """Return true if other is case-insensitive greater than or equal to self."""
        return str.__ge__(self._lowered, _try_lower(other))

    def __lt__(self, other):
        """Return true if other is case-insensitive less than self."""
        return str.__lt__(self._lowered, _try_lower(other))

    def __le__(self, other):
        """Return true if other is case-insensitive less than or equal to to self."""
        return str.__le__(self._lowered, _try_lower(other))

    def __ne__(self, other):
        """Return true if other is case-insensitive unequal to self."""
        return str.__ne__(self._lowered, _try_lower(other))


class CaseInsensitiveDict(dict):
    """Extend ``dict`` to use a case-insensitive key set."""

    def __init__(self, *args, **kwargs):
        """Create a dict with a set of lowercase keys."""
        super().__init__(*args, **kwargs)
        self._keys_to_lower()

    def __eq__(self, other):
        """Return true if other is case-insensitive equal to self."""
        return super().__eq__(CaseInsensitiveDict(other))

    def __getitem__(self, key):
        """Return value from case-insensitive lookup of ``key``."""
        return super().__getitem__(CaseInsensitiveStr(key))

    def __setitem__(self, key, value):
        """Set value with lowercase ``key``."""
        super().__setitem__(CaseInsensitiveStr(key), value)

    def __delitem__(self, key):
        """Delete value associated with case-insensitive lookup of ``key``."""
        return super().__delitem__(CaseInsensitiveStr(key))

    def __contains__(self, key):
        """Return true if key set includes case-insensitive ``key``."""
        return super().__contains__(CaseInsensitiveStr(key))

    def pop(self, key, *args, **kwargs):
        """Remove and return the value associated with case-insensitive ``key``."""
        return super().pop(CaseInsensitiveStr(key))

    def _keys_to_lower(self):
        """Convert key set to lowercase."""
        for k in list(self.keys()):
            val = super().__getitem__(k)
            super().__delitem__(k)
            self.__setitem__(CaseInsensitiveStr(k), val)


class TDSCatalog:
    """
    Parse information from a THREDDS Client Catalog.

    Attributes
    ----------
    catalog_url : str
        The url path of the catalog to parse.
    base_tds_url : str
        The top level server address
    datasets : DatasetCollection[str, Dataset]
        A dictionary of :class:`Dataset` objects, whose keys are the name of the
        dataset's name
    services : List
        A list of :class:`SimpleService` listed in the catalog
    catalog_refs : DatasetCollection[str, CatalogRef]
        A dictionary of :class:`CatalogRef` objects whose keys are the name of the
        catalog ref title.

    """

    def __init__(self, catalog_url):
        """
        Initialize the TDSCatalog object.

        Parameters
        ----------
        catalog_url : str
            The URL of a THREDDS client catalog

        """
        self.session = session_manager.create_session()

        # get catalog.xml file
        resp = self.session.get(catalog_url)
        resp.raise_for_status()

        # top level server url
        self.catalog_url = resp.url
        self.base_tds_url = _find_base_tds_url(self.catalog_url)

        # If we were given an HTML link, warn about it and try to fix to xml
        if 'html' in resp.headers['content-type']:
            import warnings
            new_url = self.catalog_url.replace('html', 'xml')
            warnings.warn(f'URL {self.catalog_url} returned HTML. Changing to: {new_url}',
                          stacklevel=2)
            self.catalog_url = new_url
            resp = self.session.get(self.catalog_url)
            resp.raise_for_status()

        # begin parsing the xml doc
        root = ET.fromstring(resp.content)
        self.catalog_name = root.attrib.get('name', 'No name found')

        self.datasets = DatasetCollection()
        self.services = []
        self.catalog_refs = DatasetCollection()
        self.metadata = {}
        self.ds_with_access_elements_to_process = []
        service_skip_count = 0
        service_skip = 0
        current_dataset = None
        previous_dataset = None
        for child in root.iter():
            tag_type = child.tag.split('}')[-1]
            if tag_type == 'dataset':
                current_dataset = child.attrib['name']
                self._process_dataset(child)

                # see if the previously processed dataset has access elements as children
                # if so, these datasets need to be processed specially when making
                # access_urls
                if previous_dataset and self.datasets[previous_dataset].access_element_info:
                    self.ds_with_access_elements_to_process.append(previous_dataset)

                previous_dataset = current_dataset

            elif tag_type == 'access':
                self.datasets[current_dataset].add_access_element_info(child)
            elif tag_type == 'catalogRef':
                self._process_catalog_ref(child)
            elif (tag_type == 'metadata') or (tag_type == ''):
                self._process_metadata(child, tag_type)
            elif tag_type == 'service':
                if (CaseInsensitiveStr(child.attrib['serviceType'])
                        != CaseInsensitiveStr('Compound')):
                    # we do not want to process single services if they
                    # are already contained within a compound service, so
                    # we need to skip over those cases.
                    if service_skip_count >= service_skip:
                        self.services.append(SimpleService(child))
                        service_skip = 0
                        service_skip_count = 0
                    else:
                        service_skip_count += 1
                else:
                    self.services.append(CompoundService(child))
                    service_skip = self.services[-1].number_of_subservices
                    service_skip_count = 0

        # Needed if the last dataset had such info, since it's only processed looking backwards
        # when a new dataset is encountered.
        if previous_dataset and self.datasets[previous_dataset].access_element_info:
            self.ds_with_access_elements_to_process.append(previous_dataset)

        self._process_datasets()

    def __str__(self):
        """Return a string representation of the catalog name."""
        return str(self.catalog_name)

    def __del__(self):
        """When TDSCatalog is deleted, close any open sessions."""
        self.session.close()

    def _process_dataset(self, element):
        catalog_url = ''
        if 'urlPath' in element.attrib and element.attrib['urlPath'] == 'latest.xml':
            catalog_url = self.catalog_url

        ds = Dataset(element, catalog_url=catalog_url)
        self.datasets[ds.name] = ds

    def _process_catalog_ref(self, element):
        catalog_ref = CatalogRef(self.catalog_url, element)
        self.catalog_refs[catalog_ref.title] = catalog_ref

    def _process_metadata(self, element, tag_type):
        if tag_type == '':
            log.warning('Trying empty tag type as metadata')
        self.metadata = TDSCatalogMetadata(element, self.metadata).metadata

    def _process_datasets(self):
        # Need to use list (of keys) because we modify the dict while iterating
        for ds_name in list(self.datasets):
            # check to see if dataset needs to have access urls created, if not,
            # remove the dataset
            has_url_path = self.datasets[ds_name].url_path is not None
            is_ds_with_access_elements_to_process = (
                ds_name in self.ds_with_access_elements_to_process
            )
            if has_url_path or is_ds_with_access_elements_to_process:
                self.datasets[ds_name].make_access_urls(
                    self.base_tds_url, self.services, metadata=self.metadata)
            else:
                self.datasets.pop(ds_name)

    @property
    def latest(self):
        """Get the latest dataset, if available."""
        for service in self.services:
            if service.is_resolver():
                latest_cat = self.catalog_url.replace('catalog.xml', 'latest.xml')
                return TDSCatalog(latest_cat).datasets[0]
        raise AttributeError('"latest" not available for this catalog')

    __repr__ = __str__


class CatalogRef:
    """
    An object for holding catalog references obtained from a THREDDS Client Catalog.

    Attributes
    ----------
    name : str
        The name of the :class:`CatalogRef` element
    href : str
        url to the :class:`CatalogRef`'s THREDDS Client Catalog
    title : str
        Title of the :class:`CatalogRef` element

    """

    def __init__(self, base_url, element_node):
        """
        Initialize the catalogRef object.

        Parameters
        ----------
        base_url : str
            URL to the base catalog that owns this reference
        element_node : :class:`~xml.etree.ElementTree.Element`
            An :class:`~xml.etree.ElementTree.Element` representing a catalogRef node

        """
        self.title = element_node.attrib['{http://www.w3.org/1999/xlink}title']
        self.name = element_node.attrib.get('name', self.title)

        # Resolve relative URLs
        href = element_node.attrib['{http://www.w3.org/1999/xlink}href']
        self.href = urljoin(base_url, href)

    def __str__(self):
        """Return a string representation of the catalog reference."""
        return str(self.title)

    def follow(self):
        """Follow the catalog reference and return a new :class:`TDSCatalog`.

        Returns
        -------
        TDSCatalog
            The referenced catalog

        """
        return TDSCatalog(self.href)

    __repr__ = __str__


class Dataset:
    """
    An object for holding Datasets obtained from a THREDDS Client Catalog.

    Attributes
    ----------
    name : str
        The name of the :class:`Dataset` element
    url_path : str
        url to the accessible dataset
    access_urls : CaseInsensitiveDict[str, str]
        A dictionary of access urls whose keywords are the access service
        types defined in the catalog (for example, "OPENDAP", "NetcdfSubset",
        "WMS", etc.

    """

    ncss_service_names = (CaseInsensitiveStr('NetcdfSubset'),
                          CaseInsensitiveStr('NetcdfServer'))

    def __init__(self, element_node, catalog_url=''):
        """Initialize the Dataset object.

        Parameters
        ----------
        element_node : :class:`~xml.etree.ElementTree.Element`
            An :class:`~xml.etree.ElementTree.Element` representing a Dataset node
        catalog_url : str
            The top level server url

        """
        self.name = element_node.attrib['name']
        self.id = element_node.attrib.get('ID', None)
        self.url_path = element_node.attrib.get('urlPath', None)
        self.catalog_name = ''
        self.access_element_info = {}
        self._resolved = False
        self._resolverUrl = None
        # if latest.xml, resolve the latest url
        if self.url_path == 'latest.xml':
            if catalog_url != '':
                self._resolved = True
                self._resolverUrl = self.url_path
                self.url_path = self.resolve_url(catalog_url)
            else:
                log.warning('Must pass along the catalog URL to resolve '
                            'the latest.xml dataset!')

    def __str__(self):
        """Return a string representation of the dataset."""
        return str(self.name)

    def resolve_url(self, catalog_url):
        """Resolve the url of the dataset when reading latest.xml.

        Parameters
        ----------
        catalog_url : str
            The catalog url to be resolved

        """
        if catalog_url != '':
            resolver_base = catalog_url.split('catalog.xml')[0]
            resolver_url = resolver_base + self.url_path
            resolver_xml = session_manager.urlopen(resolver_url)
            tree = ET.parse(resolver_xml)
            root = tree.getroot()
            self.catalog_name = root.attrib.get('name', 'No name found')
            resolved_url = ''
            found = False
            for child in root.iter():
                if not found:
                    tag_type = child.tag.split('}')[-1]
                    if tag_type == 'dataset' and 'urlPath' in child.attrib:
                        ds = Dataset(child)
                        resolved_url = ds.url_path
                        found = True
            if found:
                return resolved_url
            else:
                log.warning('no dataset url path found in latest.xml!')
        return None

    def make_access_urls(self, catalog_url, all_services, metadata=None):
        """Make fully qualified urls for the access methods enabled on the dataset.

        Parameters
        ----------
        catalog_url : str
            The top level server url
        all_services : List[SimpleService]
            list of :class:`SimpleService` objects associated with the dataset
        metadata : dict
            Metadata from the :class:`TDSCatalog`

        """
        all_service_dict = CaseInsensitiveDict({})
        for service in all_services:
            all_service_dict[service.name] = service
            if isinstance(service, CompoundService):
                for subservice in service.services:
                    all_service_dict[subservice.name] = subservice

        service_name = metadata.get('serviceName', None)

        access_urls = CaseInsensitiveDict({})
        server_url = _find_base_tds_url(catalog_url)

        # process access urls for datasets that reference top
        # level catalog services (individual or compound service
        # types).
        if service_name in all_service_dict:
            service = all_service_dict[service_name]
            if service.service_type != 'Resolver':
                # if service is a CompoundService, create access url
                # for each SimpleService
                if isinstance(service, CompoundService):
                    for subservice in service.services:
                        server_base = urljoin(server_url, subservice.base)
                        access_urls[subservice.service_type] = server_base + self.url_path
                else:
                    server_base = urljoin(server_url, service.base)
                    access_urls[service.service_type] = server_base + self.url_path

        # process access children of dataset elements
        for sname in self.access_element_info:
            if sname in all_service_dict:
                service = all_service_dict[sname]
                url_path = self.access_element_info[sname]
                server_base = urljoin(server_url, service.base)
                access_urls[service.service_type] = server_base + url_path

        self.access_urls = access_urls

    def add_access_element_info(self, access_element):
        """Create an access method from a catalog element."""
        service_name = access_element.attrib['serviceName']
        url_path = access_element.attrib['urlPath']
        self.access_element_info[service_name] = url_path

    def download(self, filename=None):
        """Download the dataset to a local file.

        Parameters
        ----------
        filename : str, optional
            The full path to which the dataset will be saved

        """
        if filename is None:
            filename = self.name
        with self.remote_open() as infile, open(filename, 'wb') as outfile:
            outfile.write(infile.read())

    def remote_open(self, mode='b', encoding='ascii', errors='ignore'):
        """Open the remote dataset for random access.

        Get a file-like object for reading from the remote dataset, providing random access,
        similar to a local file.

        Parameters
        ----------
        mode : `'b'` or `'t'`, optional
            Mode with which to open the remote data; 'b' for binary, 't' for text. Defaults
            to 'b'.

        encoding : str, optional
            If ``mode`` is text, the encoding to use to decode the binary data into text.
            Defaults to 'ascii'.

        errors : str, optional
            If ``mode`` is text, the error handling behavior to pass to `bytes.decode`.
            Defaults to 'ignore'.

        Returns
        -------
        fobj : file-like object
            A random access, file-like object for reading data

        """
        fobj = self.access_with_service('HTTPServer')
        if mode == 't':
            from io import StringIO
            fobj = StringIO(fobj.read().decode(encoding, errors))
        return fobj

    def remote_access(self, service=None, use_xarray=None):
        """Access the remote dataset.

        Open the remote dataset and get a netCDF4-compatible `Dataset` object providing
        index-based subsetting capabilities.

        Parameters
        ----------
        service : str, optional
            The name of the service to use for access to the dataset, either
            'CdmRemote' or 'OPENDAP'. Defaults to 'CdmRemote'.

        Returns
        -------
        Dataset
            Object for netCDF4-like access to the dataset

        """
        if service is None:
            service = 'CdmRemote' if 'CdmRemote' in self.access_urls else 'OPENDAP'

        if service not in (CaseInsensitiveStr('CdmRemote'), CaseInsensitiveStr('OPENDAP'),
                           CaseInsensitiveStr('DODS')):
            raise ValueError(service + ' is not a valid service for remote_access')

        return self.access_with_service(service, use_xarray)

    def subset(self, service=None):
        """Subset the dataset.

        Open the remote dataset and get a client for talking to ``service``.

        Parameters
        ----------
        service : str, optional
            The name of the service for subsetting the dataset. Defaults to 'NetcdfSubset'
            or 'NetcdfServer', in that order, depending on the services listed in the
            catalog.

        Returns
        -------
        a client for communicating using ``service``

        """
        if service is None:
            for service_name in self.ncss_service_names:
                if service_name in self.access_urls:
                    service = service_name
                    break
            else:
                raise RuntimeError('Subset access is not available for this dataset.')
        elif service not in self.ncss_service_names:
            raise ValueError(service + ' is not a valid service for subset. Options are: '
                             + ', '.join(self.ncss_service_names))

        return self.access_with_service(service)

    def access_with_service(self, service, use_xarray=None):
        """Access the dataset using a particular service.

        Return an Python object capable of communicating with the server using the particular
        service. For instance, for 'HTTPServer' this is a file-like object capable of
        HTTP communication; for OPENDAP this is a netCDF4 dataset.

        Parameters
        ----------
        service : str
            The name of the service for accessing the dataset

        Returns
        -------
            An instance appropriate for communicating using ``service``.

        """
        service = CaseInsensitiveStr(service)
        if service == 'CdmRemote':
            if use_xarray:
                from .cdmr.xarray_support import CDMRemoteStore
                try:
                    import xarray as xr
                    provider = lambda url: xr.open_dataset(CDMRemoteStore(url))  # noqa: E731
                except ImportError:
                    raise ImportError('CdmRemote access needs xarray'
                                      'to be installed.') from None
            else:
                from .cdmr import Dataset as CDMRDataset
                provider = CDMRDataset
        elif service == 'OPENDAP' or service == 'DODS':
            if use_xarray:
                try:
                    import xarray as xr
                    provider = xr.open_dataset
                except ImportError:
                    raise ImportError('xarray needs to be installed if '
                                      '`use_xarray` is True.') from None
            else:
                try:
                    from netCDF4 import Dataset as NC4Dataset
                    provider = NC4Dataset
                except ImportError:
                    raise ImportError('OPENDAP access needs netCDF4-python'
                                      'to be installed.') from None
        elif service in self.ncss_service_names:
            from .ncss import NCSS
            provider = NCSS
        elif service == 'HTTPServer' or service == CaseInsensitiveStr('http'):
            provider = session_manager.urlopen
        else:
            raise ValueError(service + ' is not an access method supported by Siphon')

        try:
            return provider(self.access_urls[service])
        except KeyError:
            raise ValueError(service + ' is not available for this dataset') from None

    __repr__ = __str__


class SimpleService:
    """Hold information about an access service enabled on a dataset.

    Attributes
    ----------
    name : str
        The name of the service
    service_type : str
        The service type (i.e. "OPENDAP", "NetcdfSubset", "WMS", etc.)
    access_urls : dict[str, str]
        A dictionary of access urls whose keywords are the access service
        types defined in the catalog (for example, "OPENDAP", "NetcdfSubset",
        "WMS", etc.)

    """

    def __init__(self, service_node):
        """Initialize the Dataset object.

        Parameters
        ----------
        service_node : :class:`~xml.etree.ElementTree.Element`
            An :class:`~xml.etree.ElementTree.Element` representing a service node

        """
        self.name = service_node.attrib['name']
        self.service_type = CaseInsensitiveStr(service_node.attrib['serviceType'])
        self.base = service_node.attrib['base']
        self.access_urls = {}

    def is_resolver(self):
        """Return whether the service is a resolver service."""
        return self.service_type == 'Resolver'


class CompoundService:
    """Hold information about compound services.

    Attributes
    ----------
    name : str
        The name of the compound service
    service_type : str
        The service type (for this object, service type will always be
        "COMPOUND")
    services : list[SimpleService]
        A list of :class:`SimpleService` objects

    """

    def __init__(self, service_node):
        """Initialize a :class:`CompoundService` object.

        Parameters
        ----------
        service_node : :class:`~xml.etree.ElementTree.Element`
            An :class:`~xml.etree.ElementTree.Element` representing a compound service node

        """
        self.name = service_node.attrib['name']
        self.service_type = CaseInsensitiveStr(service_node.attrib['serviceType'])
        self.base = service_node.attrib['base']
        services = []
        subservices = 0
        for child in list(service_node):
            services.append(SimpleService(child))
            subservices += 1

        self.services = services
        self.number_of_subservices = subservices

    def is_resolver(self):
        """Return whether the service is a resolver service.

        For a compound service, this is always False because it will never be
        a resolver.
        """
        return False


def _find_base_tds_url(catalog_url):
    """Identify the base URL of the THREDDS server from the catalog URL.

    Will retain URL scheme, host, port and username/password when present.
    """
    scheme, netloc, *_ = urlparse(catalog_url)
    return scheme + '://' + netloc


def get_latest_access_url(catalog_url, access_method):
    """Get the data access url to the latest data using a specified access method.

    These are available for a data available from a top level dataset catalog (url).
    Currently only supports the existence of one "latest" dataset.

    Parameters
    ----------
    catalog_url : str
        The URL of a top level data catalog
    access_method : str
        desired data access method (i.e. "OPENDAP", "NetcdfSubset", "WMS", etc)

    Returns
    -------
    access_url : str
        Data access URL to be used to access the latest data available from a
        given catalog using the specified `access_method`. Typically a single string,
        but not always.

    """
    return TDSCatalog(catalog_url).latest.access_urls[access_method]



# Copyright (c) 2013-2015 Siphon Contributors.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Support reading and parsing the dataset.xml documents from the netCDF Subset Service."""


import logging
import re

import numpy as np

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)


def _without_namespace(tagname):
    """Remove the xml namespace from a tag name."""
    if '}' in tagname:
        return tagname.rsplit('}', 1)[-1]
    return tagname


class _Types:
    @staticmethod
    def handle_typed_values(val, type_name, value_type):
        """Translate typed values into the appropriate python object.

        Takes an element name, value, and type and returns a list
        with the string value(s) properly converted to a python type.

        TypedValues are handled in ucar.ma2.DataType in netcdfJava
        in the DataType enum. Possibilities are:

            "boolean"
            "byte"
            "char"
            "short"
            "int"
            "long"
            "float"
            "double"
            "Sequence"
            "String"
            "Structure"
            "enum1"
            "enum2"
            "enum4"
            "opaque"
            "object"

        All of these are values written as strings in the xml, so simply
        applying int, float to the values will work in most cases (i.e.
        the TDS encodes them as string values properly).

        Example XML element:

        <attribute name="scale_factor" type="double" value="0.0010000000474974513"/>

        Parameters
        ----------
        val : string
            The string representation of the value attribute of the xml element

        type_name : string
            The string representation of the name attribute of the xml element

        value_type : string
            The string representation of the type attribute of the xml element

        Returns
        -------
        val : list
            A list containing the properly typed python values.

        """
        if value_type in ['byte', 'short', 'int', 'long']:
            try:
                if val := val.strip('[]'):
                    val = [int(v) for v in re.split('[ ,]', val) if v]
                else:
                    return [0]
            except ValueError:
                log.warning('Cannot convert "%s" to int. Keeping type as str.', val)
        elif value_type in ['float', 'double']:
            try:
                val = [float(v) for v in re.split('[ ,]', val) if v]
            except ValueError:
                log.warning('Cannot convert "%s" to float. Keeping type as str.', val)
        elif value_type == 'boolean':
            try:
                # special case for boolean type
                val = val.split()
                # values must be either true or false
                for potential_bool in val:
                    if potential_bool not in ['true', 'false']:
                        raise ValueError
                val = [item == 'true' for item in val]
            except ValueError:
                msg = 'Cannot convert values %s to boolean.'
                msg += ' Keeping type as str.'
                log.warning(msg, val)
        elif value_type == 'String':
            # nothing special for String type
            pass
        else:
            # possibilities - Sequence, Structure, enum, opaque, object,
            # and char.
            # Not sure how to handle these as I do not have an example
            # of how they would show up in dataset.xml
            log.warning('%s type %s not understood. Keeping as String.',
                        type_name, value_type)

        if not isinstance(val, list):
            val = [val]

        return val

    def handle_attribute(self, element):  # noqa
        type_name = 'attribute'
        attribute_type = None
        if 'type' in element.attrib:
            attribute_type = element.attrib['type']

        name = element.attrib['name']
        val = element.attrib['value']
        if attribute_type:
            val = self.handle_typed_values(val, type_name, attribute_type)

        return {name: val}

    def handle_values(self, element, value_type=None):  # noqa
        type_name = 'value'
        val = element.text
        if val:
            if value_type:
                val = self.handle_typed_values(val, type_name, value_type)
            else:
                val = val.split()
        else:
            increment_attrs = ['start', 'increment', 'npts']
            element_attrs = list(element.attrib)
            increment_attrs.sort()
            element_attrs.sort()
            if increment_attrs == element_attrs:
                start = float(element.attrib['start'])
                inc = float(element.attrib['increment'])
                npts = float(element.attrib['npts'])
                val = start + np.arange(npts) * inc
                val = val.tolist()

        return {'values': val}

    @staticmethod
    def handle_projectionBox(element):  # noqa
        type_name = 'projectionBox'
        pb = {}
        if element.tag == type_name:
            for child in element:
                pb[child.tag] = float(child.text)

        return {type_name: pb}

    @staticmethod
    def handle_axisRef(element):  # noqa
        return element.attrib['name']

    @staticmethod
    def handle_coordTransRef(element):  # noqa
        # type_name = "coordTransRef"
        return {'coordTransRef': element.attrib['name']}

    def handle_grid(self, element):
        grid = {}
        for attr in element.attrib:
            grid[attr] = element.attrib[attr]

        attrs = {}
        for attribute in element:
            attrs.update(self.handle_attribute(attribute))

        grid['attributes'] = attrs

        return grid

    @staticmethod
    def handle_parameter(element):
        name = element.attrib['name']
        value = element.attrib['value'].strip()
        return {name: value}

    @staticmethod
    def handle_featureDataset(element):  # noqa
        fd = {}
        for attr in element.attrib:
            fd[attr] = element.attrib[attr]
        return fd

    def handle_variable(self, element):
        return self.handle_grid(element)

    def lookup(self, handler_name):
        handler_name = 'handle_' + _without_namespace(handler_name)
        if handler_name in dir(self):
            return getattr(self, handler_name)
        else:
            log.warning('cannot find handler for element %s', handler_name)
        return None


class NCSSDataset:
    """Hold information contained in the dataset.xml NCSS document.

    In general, if a dataset.xml NCSS document is missing the information
    needed to construct an attribute, that attribute will not show up as
    part of the `NCSSDataset` object.

    Note that only gridded ncss datasets may contain the attributes
    `gridsets`, `axes`, `coordinate_transforms`, and `lat_lon_box`.

    Attributes
    ----------
    variables : dict[str, str]
        A dictionary of variables

    time_span : dict[str, datetime.datetime]
        A dictionary holding the beginning and ending iso time strings which
        define the temporal bounds of the dataset

     featureDataset : dict[str, str]
        A dictionary containing the type ["grid", "point"] and location ["url"]
        of the dataset

    accept_list : dict[str, list[str]]
        A dictionary holding the types of valid returns of the dataset by
        access method [Grid, GridAsPoint, PointFeatureCollection]

    gridsets : dict[str, set[str]]
        A dictionary of gridSets contained within the dataset

    axes : dict[str, object]
        A dictionary of coordinate axes

    coordinate_transforms : dict[str, object]
        A dictionary of coordinate transforms

    lat_lon_box : dict[str, float]
        A dictionary holding the north, south, east, and west latitude and
        longitude bounds of the dataset (in degree_east, degree_north)

    """

    def __init__(self, element):
        """Initialize a NCSSDataset object.

        Parameters
        ----------
        element : :class:`~xml.etree.ElementTree.Element`
            An :class:`~xml.etree.ElementTree.Element` representing the top level
            node of an NCSS dataset.xml doc

        """
        self._types = _Types()

        self.gridsets = {}
        self.variables = {}
        self.axes = {}
        self.coordinate_transforms = {}
        self.accept_list = {}

        self.lat_lon_box = None
        self.time_span = None
        self.featureDataset = None

        element_name = element.tag

        if element_name == 'gridDataset' or element_name == 'capabilities':

            self.featureDataset = {'type': 'grid',
                                   'url': element.attrib['location']}
            for child in element:
                self._parse_element(child)

        else:
            self._parse_element(element)

        things_to_del = []
        for thing in self.__dict__:
            if not (thing.startswith(('_', '__'))) and not getattr(self, thing):
                things_to_del.append(thing)

        for thing in things_to_del:
            delattr(self, thing)

    def _get_handler(self, handler_name):
        return self._types.lookup(handler_name)

    def _parse_element(self, element):
        element_name = element.tag

        parser = {'gridSet': self._parse_gridset, 'axis': self._parse_axis,
                  'coordTransform': self._parse_coordTransform,
                  'LatLonBox': self._parse_LatLonBox, 'TimeSpan': self._parse_TimeSpan,
                  'AcceptList': self._parse_AcceptList,
                  'featureDataset': self._parse_featureDataset,
                  'variable': self._parse_variable}

        try:
            parser[element_name](element)
        except KeyError:
            log.warning('No parser found for element %s', element_name)

    def _parse_gridset(self, element):
        element_name = element.tag
        gridset_name = element.attrib['name']
        grid_set = {}
        for child in element:
            child_name = child.tag
            handler = self._get_handler(child_name)
            if child_name in ['projectionBox', 'coordTransRef']:
                grid_set.update(handler(child))
            elif child_name in ['axisRef']:
                grid_set.setdefault(child_name, []).append(handler(child))
            elif child_name in ['grid']:
                tmp = handler(child)
                grid_name = tmp['name']
                tmp.pop('name', None)
                grid_set.setdefault(child_name, {})[grid_name] = tmp
                self.variables[grid_name] = tmp
            else:
                log.warning('Unknown child in %s: %s', element_name, child_name)
                grid_set[child.tag] = 'not handled by _parse_gridset'

        self.gridsets.update({gridset_name: grid_set})

    def _parse_axis(self, element):
        # element_name = element.tag
        axis_name = element.attrib['name']
        axis = {}
        for attr in element.attrib:
            if attr != 'name':
                axis[attr] = element.attrib[attr]
        if 'shape' in axis:
            typed_vals = self._types.handle_typed_values(axis['shape'], 'shape', 'int')
            axis['shape'] = typed_vals

        attrs = []
        for child in element:
            child_name = child.tag
            handler = self._get_handler(child_name)
            attrs.append(handler(child))

        if attrs:
            axis['attributes'] = attrs

        self.axes.update({axis_name: axis})

    def _parse_coordTransform(self, element):  # noqa
        coord_trans = {}
        name = element.attrib['name']
        for attr in element.attrib:
            if attr != 'name':
                coord_trans[attr] = element.attrib[attr]

        params = {}
        for child in element:
            child_name = child.tag
            handler = self._get_handler(child_name)
            params.update(handler(child))

        if params:
            coord_trans['parameters'] = params

        self.coordinate_transforms.update({name: coord_trans})

    def _parse_LatLonBox(self, element):  # noqa
        llb = {}
        for child in element:
            llb[child.tag] = float(child.text)
        self.lat_lon_box = llb

    def _parse_TimeSpan(self, element):  # noqa
        ts = {}
        for child in element:
            ts[child.tag] = child.text

        self.time_span = ts

    def _parse_AcceptList(self, element):  # noqa
        grid_req_types = ['Grid', 'GridAsPoint']
        # check if station (i.e.
        check = True
        grid = False
        point = False

        for child in element:
            request_type = child.tag
            if check:
                if request_type in grid_req_types:
                    grid = True
                else:
                    point = True
                check = False

            if point:
                # this is a PointFeatureCollection ncss
                return_type = child.text
                self.accept_list.setdefault('PointFeatureCollection',
                                            []).append(return_type)
            elif grid:
                # this is a grid ncss
                for grandchild in child:
                    return_type = grandchild.text
                    self.accept_list.setdefault(request_type,
                                                []).append(return_type)
            else:
                log.warning('Cannot have grid=%s and point=%s', grid, point)

    def _parse_featureDataset(self, element):  # noqa
        handler = self._get_handler(element.tag)
        self.featureDataset = handler(element)

    def _parse_variable(self, element):
        handler = self._get_handler(element.tag)
        tmp = handler(element)
        name = tmp['name']
        tmp = tmp.pop('name', None)
        self.variables[name] = tmp
